from __future__ import annotations

import logging
from dataclasses import dataclass, field, fields
from functools import cached_property, lru_cache

from typing_extensions import List, Dict, TYPE_CHECKING, Optional, Set

from .dao import AlternativeMapping
from .utils import InheritanceStrategy, module_and_class_name
from ..class_diagrams.class_diagram import (
    WrappedClass,
)
from ..class_diagrams.wrapped_field import WrappedField

if TYPE_CHECKING:
    from .ormatic import ORMatic


logger = logging.getLogger(__name__)


@dataclass
class ColumnConstructor:
    """
    Represents a column constructor that can be used to create a column in SQLAlchemy.
    """

    name: str
    """
    The name of the column.
    """

    type: str
    """
    The type of the column.
    Needs to be like "Mapped[<type>]".
    """

    constructor: Optional[str] = None
    """
    The constructor call for sqlalchemy of the column.
    """

    def __str__(self) -> str:
        if self.constructor:
            return f"{self.name}: {self.type} = {self.constructor}"
        else:
            return f"{self.name}: {self.type}"


@dataclass
class WrappedTable:
    """
    A class that wraps a dataclass and contains all the information needed to create a SQLAlchemy table from it.
    """

    wrapped_clazz: WrappedClass
    """
    The wrapped class that this table wraps.
    """

    ormatic: ORMatic
    """
    Reference to the ORMatic instance that created this WrappedTable.
    """

    builtin_columns: List[ColumnConstructor] = field(default_factory=list, init=False)
    """
    List of columns that can be directly mapped using builtin types
    """

    custom_columns: List[ColumnConstructor] = field(default_factory=list, init=False)
    """
    List for custom columns that need to by fully qualified as triple of (name, type, constructor)
    """

    foreign_keys: List[ColumnConstructor] = field(default_factory=list, init=False)
    """
    List of columns that represent foreign keys as triple of (name, type, constructor)
    """

    relationships: List[ColumnConstructor] = field(default_factory=list, init=False)
    """
    List of relationships that should be added to the table.
    """

    mapper_args: Dict[str, str] = field(default_factory=dict, init=False)

    primary_key_name: str = "database_id"
    """
    The name of the primary key column.
    """

    polymorphic_on_name: str = "polymorphic_type"
    """
    The name of the column that will be used to identify polymorphic identities if any present.
    """

    skip_fields: List[WrappedField] = field(default_factory=list)
    """
    A list of fields that should be skipped when processing the dataclass.
    """

    @property
    def primary_key(self):
        if self.parent_table is not None:
            column_type = f"ForeignKey({self.parent_table.full_primary_key_name})"
        else:
            column_type = "Integer"

        return ColumnConstructor(
            self.primary_key_name,
            f"Mapped[{module_and_class_name(int)}]",
            f"mapped_column({column_type}, primary_key=True, use_existing_column=True)",
        )

    @property
    def child_tables(self) -> List[WrappedTable]:
        return [
            self.ormatic.class_dependency_graph._dependency_graph[index]
            for index in self.ormatic.inheritance_graph.successors(
                self.wrapped_clazz.index
            )
        ]

    def create_mapper_args(self):

        # this is the root of an inheritance structure
        if self.parent_table is None and len(self.child_tables) > 0:
            self.custom_columns.append(
                (
                    ColumnConstructor(
                        self.polymorphic_on_name,
                        "Mapped[str]",
                        "mapped_column(String(255), nullable=False, use_existing_column=True)",
                    )
                )
            )
            self.mapper_args.update(
                {
                    "'polymorphic_on'": f"'{self.polymorphic_on_name}'",
                    "'polymorphic_identity'": f"'{self.tablename}'",
                }
            )

        # this inherits from something
        if self.parent_table is not None:
            self.mapper_args.update(
                {
                    "'polymorphic_identity'": f"'{self.tablename}'",
                }
            )
            # only needed for joined-table inheritance
            if self.ormatic.inheritance_strategy == InheritanceStrategy.JOINED:
                self.mapper_args.update(
                    {
                        "'inherit_condition'": f"{self.primary_key_name} == {self.parent_table.full_primary_key_name}"
                    }
                )

    @cached_property
    def full_primary_key_name(self):
        return f"{self.tablename}.{self.primary_key_name}"

    @cached_property
    def tablename(self):
        result = self.wrapped_clazz.clazz.__name__
        result += "DAO"
        return result

    @cached_property
    def parent_table(self) -> Optional[WrappedTable]:
        parents = self.ormatic.inheritance_graph.predecessors(self.wrapped_clazz.index)
        if len(parents) == 0:
            return None

        return self.ormatic.wrapped_tables[
            self.ormatic.class_dependency_graph._dependency_graph[parents[0]]
        ]

    @property
    def is_alternatively_mapped(self):
        return issubclass(self.wrapped_clazz.clazz, AlternativeMapping)

    @cached_property
    def fields(self) -> List[WrappedField]:
        """
        :return: The list of fields specified only in this associated dataclass that should be mapped.
        """

        # initialize the skip_fields of this class by excluding all parent fields
        self.skip_fields = []
        if self.parent_table is not None:
            self.skip_fields += self.parent_table.skip_fields + self.parent_table.fields

        # get all new fields given by this class
        result = [f for f in self.wrapped_clazz.fields if f not in self.skip_fields]

        # if the parent table is alternatively mapped, we need to remove the fields that are not present in the original class
        if self.parent_table is not None and self.parent_table.is_alternatively_mapped:
            # get the wrapped class of the original parent class
            og_parent_class = self.parent_table.wrapped_clazz.clazz.original_class()
            wrapped_og_parent_class = (
                self.ormatic.class_dependency_graph.get_wrapped_class(og_parent_class)
            )
            fields_in_og_class_but_not_in_dao = [
                f
                for f in wrapped_og_parent_class.fields
                if f not in self.parent_table.fields
            ]

            result = [r for r in result if r not in fields_in_og_class_but_not_in_dao]

        return result

    @lru_cache(maxsize=None)
    def parse_fields(self):

        for f in self.fields:

            logger.info("=" * 80)
            logger.info(
                f"Processing Field {self.wrapped_clazz.clazz.__name__}.{f.field.name}: {f.field.type}."
            )

            # skip private fields
            if f.field.name.startswith("_"):
                logger.info(f"Skipping since the field starts with _.")
                continue

            self.parse_field(f)

        self.create_mapper_args()

    def parse_field(self, wrapped_field: WrappedField):
        """
        Parses a given `WrappedField` and determines its type or relationship to create the
        appropriate column or define relationships in an ORM context.
        The method processes several
        types of fields, such as type types, built-in types, enumerations, one-to-one relationships,
        custom types, JSON containers, and one-to-many relationships.

        This creates the right information in the right place in the table definition to be read later by the jinja
        template.

        :param wrapped_field: An instance of `WrappedField` that contains metadata about the field
            such as its data type, whether it represents a built-in or user-defined type, or if it has
            specific ORM container properties.
        """
        if wrapped_field.is_type_type:
            logger.info(f"Parsing as type.")
            self.create_type_type_column(wrapped_field)

        elif (
            wrapped_field.is_builtin_type or wrapped_field.is_enum
        ) and not wrapped_field.is_container:
            logger.info(f"Parsing as builtin type.")
            self.create_builtin_column(wrapped_field)

        # handle one to one relationships
        elif (
            wrapped_field.is_one_to_one_relationship
            and wrapped_field.type_endpoint in self.ormatic.mapped_classes
        ):
            logger.info(f"Parsing as one to one relationship.")
            self.create_one_to_one_relationship(wrapped_field)

        # handle custom types
        elif (
            wrapped_field.is_one_to_one_relationship
            and wrapped_field.type_endpoint in self.ormatic.type_mappings
        ):
            logger.info(
                f"Parsing as custom type {self.ormatic.type_mappings[wrapped_field.resolved_type]}."
            )
            self.create_custom_type(wrapped_field)

        # handle JSON containers
        elif wrapped_field.is_collection_of_builtins:
            logger.info(f"Parsing as JSON.")
            self.create_container_of_builtins(wrapped_field)

        # handle one to many relationships
        elif wrapped_field.is_one_to_many_relationship:
            logger.info(f"Parsing as one to many relationship.")
            self.create_one_to_many_relationship(wrapped_field)
        else:
            logger.info("Skipping due to not handled type.")

    def create_builtin_column(self, wrapped_field: WrappedField):
        """
        Creates a built-in column mapping for the given wrapped field. Depending on the
        properties of the `wrapped_field`, this function determines whether it's an enum,
        a built-in type, or requires additional imports. It then constructs appropriate
        column definitions and adds them to the respective list of database mappings.

        :param wrapped_field: The WrappedField instance representing the field
            to create a built-in column for.
        """

        self.ormatic.imported_modules.add(wrapped_field.type_endpoint.__module__)
        inner_type = module_and_class_name(wrapped_field.type_endpoint)
        type_annotation = (
            f"{module_and_class_name(Optional)}[{inner_type}]"
            if wrapped_field.is_optional
            else inner_type
        )

        constructor = "mapped_column(use_existing_column=True)"
        self.builtin_columns.append(
            ColumnConstructor(
                name=wrapped_field.field.name,
                type=f"Mapped[{type_annotation}]",
                constructor=constructor,
            )
        )

    def create_type_type_column(self, wrapped_field: WrappedField):
        """
        Create a column for a field of type `Type`.
        :param wrapped_field: The field to extract type information from.
        :return:
        """
        column_name = wrapped_field.field.name
        column_type = (
            f"Mapped[TypeType]"
            if not wrapped_field.is_optional
            else f"Mapped[{module_and_class_name(Optional)}[TypeType]]"
        )
        column_constructor = f"mapped_column(TypeType, nullable={wrapped_field.is_optional}, use_existing_column=True)"
        self.custom_columns.append(
            ColumnConstructor(column_name, column_type, column_constructor)
        )

    def create_one_to_one_relationship(self, wrapped_field: WrappedField):
        """
        Create a one-to-one relationship with using the given field.
        This adds a foreign key and a relationship to this table.

        :param wrapped_field: The field to get the information from.
        """
        # create foreign key
        fk_name = f"{wrapped_field.field.name}{self.ormatic.foreign_key_postfix}"
        fk_type = (
            f"Mapped[{module_and_class_name(Optional)}[{module_and_class_name(int)}]]"
            if wrapped_field.is_optional
            else "Mapped[int]"
        )

        # get the target table
        target_wrapped_table = self.ormatic.wrapped_tables[
            self.ormatic.class_dependency_graph.get_wrapped_class(
                wrapped_field.type_endpoint
            )
        ]

        # columns have to be nullable and use_alter=True since the insertion order might be incorrect otherwise
        fk_column_constructor = f"mapped_column(ForeignKey('{target_wrapped_table.full_primary_key_name}', use_alter=True), nullable=True, use_existing_column=True)"

        self.foreign_keys.append(
            ColumnConstructor(fk_name, fk_type, fk_column_constructor)
        )

        # create relationship to remote side
        rel_name = f"{wrapped_field.field.name}"
        rel_type = f"Mapped[{target_wrapped_table.tablename}]"
        # relationships have to be post updated since since it won't work in the case of subclasses with another ref otherwise
        rel_constructor = f"relationship('{target_wrapped_table.tablename}', uselist=False, foreign_keys=[{fk_name}], post_update=True)"
        self.relationships.append(
            ColumnConstructor(rel_name, rel_type, rel_constructor)
        )

    def create_one_to_many_relationship(self, wrapped_field: WrappedField):
        """
        Creates a one-to-many relationship mapping for the given wrapped field.
        The target side of the wrapped field gets a foreign key to this table with a unique name.
        This table gets a relationship that joins the target table with the foreign key.

        :param wrapped_field: The field for the one-to-many relationship.
        """

        # get the target table
        target_wrapped_table = self.ormatic.wrapped_tables[
            self.ormatic.class_dependency_graph.get_wrapped_class(
                wrapped_field.type_endpoint
            )
        ]

        # create a foreign key to this on the remote side
        fk_name = f"{self.tablename.lower()}_{wrapped_field.field.name}{self.ormatic.foreign_key_postfix}"
        fk_type = (
            f"Mapped[{module_and_class_name(Optional)}[{module_and_class_name(int)}]]"
        )
        fk_column_constructor = f"mapped_column(ForeignKey('{self.full_primary_key_name}', use_alter=True), nullable=True, use_existing_column=True)"
        target_wrapped_table.foreign_keys.append(
            ColumnConstructor(fk_name, fk_type, fk_column_constructor)
        )

        # create a relationship with a list to collect the other side
        rel_name = f"{wrapped_field.field.name}"
        rel_type = (
            f"Mapped[{module_and_class_name(List)}[{target_wrapped_table.tablename}]]"
        )
        rel_constructor = f"relationship('{target_wrapped_table.tablename}', foreign_keys='[{target_wrapped_table.tablename}.{fk_name}]', post_update=True)"
        self.relationships.append(
            ColumnConstructor(rel_name, rel_type, rel_constructor)
        )

    def create_container_of_builtins(self, wrapped_field: WrappedField):
        """
        Create a column for a list-like of built-in values.

        :param wrapped_field: The field to extract the information from.
        """
        self.ormatic.imported_modules.add("typing_extensions")
        column_name = wrapped_field.field.name
        container = Set if issubclass(wrapped_field.container_type, set) else List
        column_type = f"Mapped[{module_and_class_name(container)}[{module_and_class_name(wrapped_field.type_endpoint)}]]"
        column_constructor = f"mapped_column(JSON, nullable={wrapped_field.is_optional}, use_existing_column=True)"
        self.custom_columns.append(
            ColumnConstructor(column_name, column_type, column_constructor)
        )

    def create_custom_type(self, wrapped_field: WrappedField):
        custom_type = self.ormatic.type_mappings[wrapped_field.type_endpoint]
        column_name = wrapped_field.field.name
        column_type = (
            f"Mapped[{custom_type.__module__}.{custom_type.__name__}]"
            if not wrapped_field.is_optional
            else f"Mapped[{module_and_class_name(Optional)}[{custom_type.__module__}.{custom_type.__name__}]]"
        )

        constructor = f"mapped_column({custom_type.__module__}.{custom_type.__name__}, nullable={wrapped_field.is_optional}, use_existing_column=True)"

        self.custom_columns.append(
            ColumnConstructor(column_name, column_type, constructor)
        )

    @property
    def base_class_name(self):
        if self.parent_table is not None:
            return self.parent_table.tablename
        else:
            return "Base"

    def __hash__(self):
        return hash(self.wrapped_clazz)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Set, Optional
from typing import TextIO

import rustworkx as rx

from sqlalchemy import TypeDecorator
from typing_extensions import List, Type, Dict
from sortedcontainers import SortedSet

from .custom_types import TypeType
from .dao import AlternativeMapping

from .sqlalchemy_generator import SQLAlchemyGenerator
from .utils import InheritanceStrategy, module_and_class_name
from .wrapped_table import WrappedTable
from ..class_diagrams.class_diagram import (
    ClassDiagram,
    Relation,
    WrappedClass,
)
from ..class_diagrams.wrapped_field import WrappedField

logger = logging.getLogger(__name__)


class AlternativelyMaps(Relation):
    """
    Edge type that says that the source alternativly maps the target, e. g.
    `AlternativeMaps(source=PointMapping, target=Point)` means that PointMapping is the mapping for Point.
    """


@dataclass
class ORMatic:
    """
    ORMatic is a tool for generating SQLAlchemy ORM models from a set of dataclasses.
    """

    class_dependency_graph: ClassDiagram
    """
    The class diagram to add the orm for.
    """

    alternative_mappings: List[Type[AlternativeMapping]] = field(default_factory=list)
    """
    List of alternative mappings that should be used to map classes.
    """

    type_mappings: Dict[Type, Type[TypeDecorator]] = field(default_factory=dict)
    """
    A dict that maps classes to custom types that should be used to save the classes.
    They keys of the type mappings must be disjoint with the classes given..
    """

    inheritance_strategy: InheritanceStrategy = InheritanceStrategy.JOINED
    """
    The inheritance strategy to use.
    """

    foreign_key_postfix = "_id"
    """
    The postfix that will be added to foreign key columns (not the relationships).
    """

    imported_modules: SortedSet[str] = field(default_factory=SortedSet, init=False)
    """
    A set of modules that need to be imported.
    """

    type_annotation_map: Dict[str, str] = field(default_factory=dict, init=False)
    """
    The string version of type mappings that is used in jinja.
    """

    inheritance_graph: rx.PyDiGraph[int] = field(default=None, init=False)
    """
    A graph that represents the inheritance structure of the classes. Extracted from the class dependency graph.
    """

    wrapped_tables: Dict[WrappedClass, WrappedTable] = field(
        default_factory=dict, init=False
    )
    """
    The wrapped tables instances for the SQLAlchemy conversion.
    """

    def __post_init__(self):
        self.type_mappings[Type] = TypeType
        self.imported_modules.add(Type.__module__)
        self._create_inheritance_graph()
        self._add_alternative_mappings_to_class_diagram()
        self._create_wrapped_tables()

        for wrapped_table in self.wrapped_tables.values():
            self.imported_modules.add(wrapped_table.wrapped_clazz.clazz.__module__)

    def _create_wrapped_tables(self):
        for wrapped_clazz in self.wrapped_classes_in_topological_order:

            # check if the class has an alternative mapping
            if alternative_mapping := self.get_alternative_mapping(wrapped_clazz):
                # add the alternative mapping
                self.wrapped_tables[wrapped_clazz] = WrappedTable(
                    wrapped_clazz=alternative_mapping, ormatic=self
                )
            else:
                # add the class normally
                self.wrapped_tables[wrapped_clazz] = WrappedTable(
                    wrapped_clazz=wrapped_clazz, ormatic=self
                )

    def _create_inheritance_graph(self):
        self.inheritance_graph = rx.PyDiGraph()
        self.inheritance_graph.add_nodes_from(
            [w.index for w in self.class_dependency_graph.wrapped_classes]
        )
        for edge in self.class_dependency_graph.inheritance_relations:
            self.inheritance_graph.add_edge(edge.source.index, edge.target.index, None)

    def _add_alternative_mappings_to_class_diagram(self):
        """
        Add alternative mappings to the class diagram.
        """
        for alternative_mapping in self.alternative_mappings:
            wrapped_alternative_mapping = WrappedClass(clazz=alternative_mapping)
            self.class_dependency_graph.add_node(wrapped_alternative_mapping)
            self.class_dependency_graph.add_relation(
                AlternativelyMaps(
                    source=wrapped_alternative_mapping,
                    target=self.class_dependency_graph.get_wrapped_class(
                        alternative_mapping.original_class()
                    ),
                )
            )

    @property
    def alternatively_maps_relations(self) -> List[AlternativelyMaps]:
        return [
            edge
            for edge in self.class_dependency_graph._dependency_graph.edges()
            if isinstance(edge, AlternativelyMaps)
        ]

    def get_alternative_mapping(
        self, wrapped_class: WrappedClass
    ) -> Optional[WrappedClass]:
        """
        Finds and returns an alternative mapping for the given wrapped class,
        if one exists, based on the relations specified in
        `alternatively_maps_relations`.

        :param wrapped_class: The wrapped class for which an alternative
            mapping is to be searched.
        :return: An alternate mapping of the type WrappedClass if found,
            otherwise None.
        """
        for rel in self.alternatively_maps_relations:
            if rel.target == wrapped_class:
                return rel.source
        return None

    def create_type_annotations_map(self):
        self.type_annotation_map = {"Type": "TypeType"}
        for clazz, custom_type in self.type_mappings.items():
            self.type_annotation_map[module_and_class_name(clazz)] = (
                module_and_class_name(custom_type)
            )

    @property
    def wrapped_classes_in_topological_order(self) -> List[WrappedClass]:
        """
        :return: List of all tables in topological order.
        """
        return [
            self.class_dependency_graph._dependency_graph[index]
            for index in rx.topological_sort(self.inheritance_graph)
        ]

    @property
    def mapped_classes(self) -> List[Type]:
        return [key.clazz for key in self.wrapped_tables.keys()]

    def make_all_tables(self):
        for table in self.wrapped_tables.values():
            table.parse_fields()

    def foreign_key_name(self, wrapped_field: WrappedField) -> str:
        """
        :return: A foreign key name for the given field.
        """
        return f"{wrapped_field.clazz.clazz.__name__.lower()}_{wrapped_field.field.name}{self.foreign_key_postfix}"

    def to_sqlalchemy_file(self, file: TextIO):
        """
        Generate a Python file with SQLAlchemy declarative mappings from the ORMatic models.

        :param file: The file to write to
        """
        sqlalchemy_generator = SQLAlchemyGenerator(self)
        sqlalchemy_generator.to_sqlalchemy_file(file)



from __future__ import annotations

from dataclasses import dataclass, field
from typing_extensions import ClassVar, Optional, List, Any

try:
    from rustworkx_utils import GraphVisualizer
except ImportError:
    GraphVisualizer = None

import rustworkx as rx


@dataclass
class ColorLegend:
    name: str = field(default="Other")
    color: str = field(default="white")


# ---- rustworkx-backed node wrapper to mimic needed anytree.Node API ----
@dataclass
class RWXNode:
    name: str
    weight: str = field(default='')
    data: Optional[Any] = field(default=None)
    _primary_parent_id: Optional[int] = None
    color: ColorLegend = field(default_factory=ColorLegend)
    # Grouping/boxing options
    wrap_subtree: bool = field(default=False)
    wrap_facecolor: Optional[str] = field(default=None)
    wrap_edgecolor: Optional[str] = field(default=None)
    wrap_alpha: float = field(default=0.08)
    # Visual emphasis options
    enclosed: bool = field(default=False)
    id_: int = field(init=False)
    _graph: ClassVar[rx.PyDAG] = rx.PyDAG()
    enclosed_name: ClassVar[str] = 'enclosed'

    def __post_init__(self):
        # store self as node data to keep a 1:1 mapping
        self.id: int = self._graph.add_node(self)

    # Non-primary connect: add edge without changing primary parent pointer
    def add_parent(self, parent: "RWXNode", edge_weight=None):
        # Avoid self-loops
        if parent is self:
            return
        # Do not add duplicate edges between the same two nodes
        if self._graph.has_edge(parent.id, self.id):
            return
        # Avoid creating cycles: PyDAG will raise if creates a cycle
        self._graph.add_edge(parent.id, self.id, edge_weight if edge_weight is not None else self.weight)

    def remove(self):
        self._graph.remove_node(self.id)

    def remove_node(self, node: RWXNode):
        self._graph.remove_node(node.id)

    def remove_child(self, child: RWXNode):
        child.remove_parent(self)

    def remove_parent(self, parent: RWXNode):
        self._graph.remove_edge(parent.id, self.id)

    @property
    def ancestors(self) -> List[RWXNode]:
        node_ids = rx.ancestors(self._graph, self.id)
        return [self._graph[n_id] for n_id in node_ids]

    @property
    def parents(self) -> List[RWXNode]:
        # In this environment rustworkx returns node data objects directly
        return self._graph.predecessors(self.id)

    @property
    def parent(self) -> Optional["RWXNode"]:
        if self._primary_parent_id is None:
            return None
        return self._graph[self._primary_parent_id]

    @parent.setter
    def parent(self, value: Optional["RWXNode"]):
        if value is None:
            # detach current parent
            self._graph.remove_edge(self._primary_parent_id, self.id)
            self._primary_parent_id = None
            return
        # Create edge and set as primary (no need to detach non-primary edges)
        self.add_parent(value)
        self._primary_parent_id = value.id

    @property
    def children(self) -> List["RWXNode"]:
        # In this environment rustworkx returns node data objects directly
        return self._graph.successors(self.id)

    @property
    def descendants(self) -> List["RWXNode"]:
        desc_ids = rx.descendants(self._graph, self.id)
        return [self._graph[nid] for nid in desc_ids]

    @property
    def leaves(self) -> List["RWXNode"]:
        return [n for n in [self] + self.descendants if self._graph.out_degree(n.id) == 0]

    @property
    def root(self) -> "RWXNode":
        n = self
        while n.parent is not None:
            n = n.parent
        return n

    def __str__(self):
        return self.name

    def visualize(self, figsize=(35, 30), node_size=7000, font_size=25, spacing_x: float = 4, spacing_y: float = 4,
                  curve_scale: float = 0.5, layout: str = 'tidy', edge_style: str = 'orthogonal',
                  label_max_chars_per_line: Optional[int] = 13):
        """Render a rooted, top-to-bottom directed graph.
        Delegates to a dedicated visualizer class to keep this method small and reusable.
        """
        if not GraphVisualizer:
            raise RuntimeError("rustworkx_utils is not installed. Please install it with `pip install rustworkx_utils`")
        visualizer = GraphVisualizer(
            node=self,
            figsize=figsize,
            node_size=node_size,
            font_size=font_size,
            spacing_x=spacing_x,
            spacing_y=spacing_y,
            curve_scale=curve_scale,
            layout=layout,
            edge_style=edge_style,
            label_max_chars_per_line=label_max_chars_per_line,
        )
        return visualizer.render()



from dataclasses import dataclass
from typing_extensions import List

from ..predicate import Predicate
from ..symbol_graph import SymbolGraph, WrappedInstance, PredicateRelation
from ...ormatic.dao import AlternativeMapping, T


@dataclass
class SymbolGraphMapping(AlternativeMapping[SymbolGraph]):
    """
    Mapping specific for SymbolGraph.
    Import this class when you want to persist SymbolGraph.
    """

    instances: List[WrappedInstance]

    predicate_relations: List[PredicateRelation]

    @classmethod
    def create_instance(cls, obj: SymbolGraph):
        return cls(
            instances=obj.wrapped_instances,
            predicate_relations=list(obj.relations()),
        )

    def create_from_dao(self) -> T:
        result = Predicate.build_symbol_graph()
        for instance in self.instances:
            result.add_instance(instance)
        for relation in self.predicate_relations:
            result.add_relation(relation)
        return result



from __future__ import annotations

from collections import UserDict, defaultdict
from contextlib import contextmanager
from copy import copy


from . import logger
from .enums import EQLMode, PredicateType
from .rxnode import RWXNode, ColorLegend

"""
Core symbolic expression system used to build and evaluate entity queries.

This module defines the symbolic types (variables, sources, logical and
comparison operators) and the evaluation mechanics.
"""
import contextvars
import operator
import typing
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from functools import lru_cache

from typing_extensions import (
    Iterable,
    Any,
    Optional,
    Type,
    Dict,
    ClassVar,
    Union as TypingUnion,
    Generic,
    TypeVar,
    TYPE_CHECKING,
    Set,
)
from typing_extensions import List, Tuple, Callable


from .cache_data import (
    cache_enter_count,
    cache_search_count,
    cache_match_count,
    is_caching_enabled,
    SeenSet,
    IndexedCache,
    get_cache_keys_for_class_,
    yield_class_values_from_cache,
)
from .failures import MultipleSolutionFound, NoSolutionFound
from .utils import IDGenerator, is_iterable, generate_combinations, lazy_iterate_dicts
from .hashed_data import HashedValue, HashedIterable, T

if TYPE_CHECKING:
    from .conclusion import Conclusion

_symbolic_mode = contextvars.ContextVar("symbolic_mode", default=None)


def _set_symbolic_mode(mode: EQLMode):
    """
    Set symbolic construction mode.

    :param mode: Can be Query or Rule.
    """
    _symbolic_mode.set(mode)


def in_symbolic_mode(mode: Optional[EQLMode] = None) -> bool:
    """
    Check whether symbolic construction mode is currently active.

    :returns: True if symbolic mode is enabled, otherwise False.
    """
    current_mode = _symbolic_mode.get()
    return current_mode == mode if mode else current_mode is not None


T = TypeVar("T")

id_generator = IDGenerator()

RWXNode.enclosed_name = "Selected Variable"


@dataclass(eq=False)
class SymbolicExpression(Generic[T], ABC):
    """
    Base class for all symbolic expressions.

    Symbolic expressions form a tree and are evaluated lazily to produce
    bindings for variables, subject to logical constraints.

    :ivar _child_: Optional child expression.
    :ivar _id_: Unique identifier of this node.
    :ivar _node_: Backing anytree.Node for visualization and traversal.
    :ivar _conclusion_: Set of conclusion actions attached to this node.
    :ivar _yield_when_false__: If True, may yield even when the expression is false.
    :ivar _is_false_: Internal flag indicating evaluation result for this node.
    """

    _child_: Optional[SymbolicExpression] = field(init=False)
    _id_: int = field(init=False, repr=False, default=None)
    _node_: RWXNode = field(init=False, default=None, repr=False)
    _id_expression_map_: ClassVar[Dict[int, SymbolicExpression]] = {}
    _conclusion_: typing.Set[Conclusion] = field(init=False, default_factory=set)
    _symbolic_expression_stack_: ClassVar[List[SymbolicExpression]] = []
    _yield_when_false_: bool = field(init=False, repr=False, default=False)
    _is_false_: bool = field(init=False, repr=False, default=False)
    _seen_parent_values_: Dict[bool, SeenSet] = field(
        default_factory=lambda: {True: SeenSet(), False: SeenSet()},
        init=False,
        repr=False,
    )
    _seen_parent_values_by_parent_: Dict[int, Dict[bool, SeenSet]] = field(
        default_factory=dict, init=False, repr=False
    )
    _eval_parent_: Optional[SymbolicExpression] = field(
        default=None, init=False, repr=False
    )
    _plot_color__: Optional[ColorLegend] = field(default=None, init=False, repr=False)

    def __post_init__(self):
        if not self._id_:
            self._id_ = id_generator(self)
            self._create_node_()
            self._id_expression_map_[self._id_] = self
        if hasattr(self, "_child_") and self._child_ is not None:
            self._update_child_()

    def _update_child_(self, child: Optional[SymbolicExpression] = None):
        child = child or self._child_
        self._child_ = self._update_children_(child)[0]

    def _update_children_(
        self, *children: SymbolicExpression
    ) -> Tuple[SymbolicExpression, ...]:
        children: Dict[int, SymbolicExpression] = dict(enumerate(children))
        for k, v in children.items():
            if not isinstance(v, SymbolicExpression):
                children[k] = Literal(v)
        for k, v in children.items():
            # With graph structure, do not copy nodes; just connect an edge.
            v._node_.parent = self._node_
        return tuple(children.values())

    def _create_node_(self):
        self._node_ = RWXNode(self._name_, data=self, color=self._plot_color_)

    def _reset_cache_(self) -> None:
        """
        Reset the cache of the symbolic expression and its children.
        """
        self._reset_only_my_cache_()
        for child in self._children_:
            child._reset_cache_()

    def _reset_only_my_cache_(self) -> None:
        """
        Reset only the cache of this symbolic expression.
        """
        self._seen_parent_values_ = {True: SeenSet(), False: SeenSet()}
        # Also reset per-parent duplicate tracking and runtime eval parent to ensure reevaluation works
        self._seen_parent_values_by_parent_ = {}
        self._eval_parent_ = None

    @abstractmethod
    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        """
        Evaluate the symbolic expression and set the operands indices.
        This method should be implemented by subclasses.
        """
        pass

    def _add_conclusion_(self, conclusion: Conclusion):
        self._conclusion_.add(conclusion)

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(
        self, child: Optional[SymbolicExpression] = None, when_true: bool = True
    ):
        if self._parent_:
            vars = self._parent_._required_variables_from_child_(
                self, when_true=when_true
            )
        else:
            vars = HashedIterable()
        if when_true or (when_true is None):
            for child in self._children_:
                for conc in child._conclusion_:
                    vars.update(conc._unique_variables_)
        return vars

    @property
    def _conclusions_of_all_descendants_(self) -> List[Conclusion]:
        return [conc for child in self._descendants_ for conc in child._conclusion_]

    @property
    def _parent_(self) -> Optional[SymbolicExpression]:
        if self._eval_parent_ is not None:
            return self._eval_parent_
        elif self._node_.parent is not None:
            return self._node_.parent.data
        return None

    @_parent_.setter
    def _parent_(self, value: Optional[SymbolicExpression]):
        self._node_.parent = value._node_ if value is not None else None
        if value is not None and hasattr(value, "_child_"):
            value._child_ = self

    @property
    def _conditions_root_(self) -> SymbolicExpression:
        """
        Get the root of the symbolic expression tree that contains conditions.
        """
        conditions_root = self._root_
        while conditions_root._child_ is not None:
            conditions_root = conditions_root._child_
            if isinstance(conditions_root._parent_, Entity):
                break
        return conditions_root

    @property
    def _root_(self) -> SymbolicExpression:
        """
        Get the root of the symbolic expression tree.
        """
        return self._node_.root.data

    @property
    @abstractmethod
    def _name_(self) -> str:
        pass

    @property
    def _all_nodes_(self) -> List[SymbolicExpression]:
        return [self] + self._descendants_

    @property
    def _all_node_names_(self) -> List[str]:
        return [node._node_.name for node in self._all_nodes_]

    @property
    def _descendants_(self) -> List[SymbolicExpression]:
        return [d.data for d in self._node_.descendants]

    @property
    def _children_(self) -> List[SymbolicExpression]:
        return [c.data for c in self._node_.children]

    @classmethod
    def _current_parent_(cls) -> Optional[SymbolicExpression]:
        if cls._symbolic_expression_stack_:
            return cls._symbolic_expression_stack_[-1]
        return None

    @property
    def _sources_(self):
        vars = [v.data for v in self._node_.leaves]
        while any(isinstance(v, SymbolicExpression) for v in vars):
            vars = {
                (
                    v._domain_source_.domain
                    if isinstance(v, Variable) and v._domain_source_
                    else v
                )
                for v in vars
            }
            for v in copy(vars):
                if isinstance(v, SymbolicExpression):
                    vars.remove(v)
                    vars.update(set(v._all_variable_instances_))
        sources = set(HashedIterable(vars))
        return sources

    @property
    @lru_cache(maxsize=None)
    def _unique_variables_(self) -> HashedIterable[Variable]:
        unique_variables = HashedIterable()
        for var in self._all_variable_instances_:
            unique_variables.add(var)
        return unique_variables

    @property
    @abstractmethod
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        """
        Get the leaf instances of the symbolic expression.
        This is useful for accessing the leaves of the symbolic expression tree.
        """
        ...

    def _is_duplicate_output_(self, output: Dict[int, HashedValue]) -> bool:
        required_vars = self._parent_._required_variables_from_child_(
            self, when_true=not self._is_false_
        )
        if not required_vars:
            return False
        required_output = {k: v for k, v in output.items() if k in required_vars}
        if not required_output:
            return False
        # Use a per-parent seen set to avoid suppressing outputs across different parent contexts
        parent_id = self._parent_._id_
        seen_by_truth = self._seen_parent_values_by_parent_.setdefault(
            parent_id, {True: SeenSet(), False: SeenSet()}
        )
        seen_set = seen_by_truth[not self._is_false_]
        if seen_set.check(required_output):
            return True
        else:
            seen_set.add(required_output)
            return False

    @property
    def _plot_color_(self) -> ColorLegend:
        return self._plot_color__

    @_plot_color_.setter
    def _plot_color_(self, value: ColorLegend):
        self._plot_color__ = value
        self._node_.color = value

    def __and__(self, other):
        return AND(self, other)

    def __or__(self, other):
        return _optimize_or(self, other)

    def __invert__(self):
        return Not(self)

    def __enter__(self, in_rule_mode: bool = False):
        node = self
        to_return = self
        if in_rule_mode or in_symbolic_mode(EQLMode.Rule):
            if (node is self._root_) or (node._parent_ is self._root_):
                node = node._conditions_root_
        if isinstance(node, Variable) and node._parent_ is None:
            node = An(Entity(selected_variables=[node]))
            to_return = node
        SymbolicExpression._symbolic_expression_stack_.append(node)
        return to_return

    def __exit__(self, *args):
        SymbolicExpression._symbolic_expression_stack_.pop()

    def __hash__(self):
        return hash(id(self))

    def __repr__(self):
        return self._name_


@dataclass(eq=False, repr=False)
class CanBehaveLikeAVariable(SymbolicExpression[T], ABC):
    """
    This class adds the monitoring/tracking behaviour on variables that tracks attribute access, calling,
    and comparison operations.
    """

    _var_: CanBehaveLikeAVariable[T] = field(init=False, default=None)
    """
    A variable that is used if the child class to this class want to provide a variable to be tracked other than 
    itself, this is specially useful for child classes that holds a variable instead of being a variable and want
     to delegate the variable behaviour to the variable it has instead.
    For example, this is the case for the ResultQuantifiers & QueryDescriptors that operate on a single selected
    variable.
    """

    def __getattr__(self, name: str) -> CanBehaveLikeAVariable[T]:
        # Prevent debugger/private attribute lookups from being interpreted as symbolic attributes
        if not in_symbolic_mode():
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute {name}, maybe you forgot to "
                f"use the symbolic_mode context manager?"
            )
        if name.startswith("__") and name.endswith("__"):
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute {name}"
            )
        return Attribute(self, name)

    def __getitem__(self, key) -> CanBehaveLikeAVariable[T]:
        self._if_not_in_symbolic_mode_raise_error_("__getitem__")
        return Index(self, key)

    def __call__(self, *args, **kwargs) -> CanBehaveLikeAVariable[T]:
        self._if_not_in_symbolic_mode_raise_error_("__call__")
        return Call(self, args, kwargs)

    def __eq__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__eq__")
        return Comparator(self, other, operator.eq)

    def __contains__(self, item) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__contains__")
        return Comparator(item, self, operator.contains)

    def __ne__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__ne__")
        return Comparator(self, other, operator.ne)

    def __lt__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__lt__")
        return Comparator(self, other, operator.lt)

    def __le__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__le__")
        return Comparator(self, other, operator.le)

    def __gt__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__gt__")
        return Comparator(self, other, operator.gt)

    def __ge__(self, other) -> Comparator:
        self._if_not_in_symbolic_mode_raise_error_("__ge__")
        return Comparator(self, other, operator.ge)

    def _if_not_in_symbolic_mode_raise_error_(self, method_name: str) -> None:
        if not in_symbolic_mode():
            raise AttributeError(
                f"You are not in symbolic_mode {self.__class__.__name__} object has no attribute"
                f" {method_name}"
            )

    def __hash__(self):
        return super().__hash__()


@dataclass(eq=False)
class ResultQuantifier(CanBehaveLikeAVariable[T], ABC):
    """
    Base for quantifiers that return concrete results from entity/set queries
    (e.g., An, The).
    """

    _child_: QueryObjectDescriptor[T]

    def __post_init__(self):
        super().__post_init__()
        self._var_ = self._child_._var_

    @property
    def _name_(self) -> str:
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def evaluate(
        self,
    ) -> TypingUnion[Iterable[T], T, Iterable[Dict[SymbolicExpression[T], T]]]:
        """
        This is the method called by the user to evaluate the full query.
        """
        ...

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(
        self, child: Optional[SymbolicExpression] = None, when_true: bool = True
    ):
        child = self._child_ if child is None else child
        if self._parent_:
            vars = self._parent_._required_variables_from_child_(
                self, when_true=when_true
            )
        else:
            vars = HashedIterable()
        for var in child.selected_variables:
            vars.add(var)
            vars.update(var._unique_variables_)
        if when_true or (when_true is None):
            for conc in child._conclusion_:
                vars.update(conc._unique_variables_)
        return vars

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        return self._child_._all_variable_instances_

    def _process_result_(
        self, result: Dict[int, HashedValue]
    ) -> TypingUnion[T, UnificationDict]:
        if isinstance(self._child_, Entity):
            return result[self._child_.selected_variable._id_].value
        elif isinstance(self._child_, SetOf):
            selected_variables_ids = [v._id_ for v in self._child_.selected_variables]
            return UnificationDict(
                {
                    self._id_expression_map_[var_id]: value
                    for var_id, value in result.items()
                    if var_id in selected_variables_ids
                }
            )
        else:
            raise NotImplementedError(f"Unknown child type {type(self._child_)}")

    def visualize(
        self,
        figsize=(35, 30),
        node_size=7000,
        font_size=25,
        spacing_x: float = 4,
        spacing_y: float = 4,
        layout: str = "tidy",
        edge_style: str = "orthogonal",
        label_max_chars_per_line: Optional[int] = 13,
    ):
        """
        Visualize the query graph, for arguments' documentation see `rustworkx_utils.RWXNode.visualize`.
        """
        self._node_.visualize(
            figsize=figsize,
            node_size=node_size,
            font_size=font_size,
            spacing_x=spacing_x,
            spacing_y=spacing_y,
            layout=layout,
            edge_style=edge_style,
            label_max_chars_per_line=label_max_chars_per_line,
        )

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("ResultQuantifier", "#9467bd")

    @_plot_color_.setter
    def _plot_color_(self, value: ColorLegend):
        self._plot_color__ = value
        self._node_.color = value


class UnificationDict(UserDict):
    """
    A dictionary which maps all expressions that are on a single variable to the original variable id.
    """

    def __getitem__(self, key: CanBehaveLikeAVariable[T]) -> T:
        key = key._id_expression_map_[key._var_._id_]
        return super().__getitem__(key).value


@dataclass(eq=False)
class The(ResultQuantifier[T]):
    """
    Quantifier that expects exactly one result; raises MultipleSolutionFound if more.
    """

    def evaluate(self) -> TypingUnion[Iterable[T], T, UnificationDict]:
        result = self._evaluate_()
        result = self._process_result_(result)
        self._reset_cache_()
        return result

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        v = self._evaluate_(sources, yield_when_false=yield_when_false)
        yield v
        return

    def _evaluate_(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Dict[int, HashedValue]:
        sources = sources or {}
        self._yield_when_false_ = yield_when_false
        self._child_._eval_parent_ = self
        if self._id_ in sources:
            return sources
        sol_gen = self._child_._evaluate__(sources)
        result = None
        for sol in sol_gen:
            if result is None:
                result = sol
                result.update(sources)
            else:
                raise MultipleSolutionFound(result, sol)
        if result is None:
            self._is_false_ = True
        if self._is_false_:
            if self._yield_when_false_:
                result = sources
            else:
                raise NoSolutionFound(self._child_)
        else:
            result[self._id_] = result[self._var_._id_]
        return result


@dataclass(eq=False, repr=False)
class An(ResultQuantifier[T]):
    """Quantifier that yields all matching results one by one."""

    def __post_init__(self):
        super().__post_init__()
        self._node_.wrap_subtree = True

    def evaluate(
        self,
    ) -> Iterable[TypingUnion[T, Dict[TypingUnion[T, SymbolicExpression[T]], T]]]:
        with symbolic_mode(mode=None):
            results = self._evaluate__()
            assert not in_symbolic_mode()
            yield from map(self._process_result_, results)
        self._reset_cache_()

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[T]:
        sources = sources or {}
        if self._id_ in sources:
            if self is self._conditions_root_ or isinstance(
                self._parent_, LogicalOperator
            ):
                original_me = self._id_expression_map_[self._id_]
                self._is_false_ = original_me._is_false_
                if not original_me._is_false_ or yield_when_false:
                    yield sources
            else:
                yield sources
        else:
            self._yield_when_false_ = yield_when_false
            any_yielded = False
            self._child_._eval_parent_ = self
            values = self._child_._evaluate__(
                sources, yield_when_false=self._yield_when_false_
            )
            for value in values:
                any_yielded = True
                self._is_false_ = self._child_._is_false_
                if self._yield_when_false_ or not self._is_false_:
                    value.update(sources)
                    if self._var_:
                        value.update({self._id_: value[self._var_._id_]})
                    yield value


@dataclass(eq=False)
class QueryObjectDescriptor(CanBehaveLikeAVariable[T], ABC):
    """
    Describes the queried object(s), could be a query over a single variable or a set of variables,
    also describes the condition(s)/properties of the queried object(s).
    """

    _child_: Optional[SymbolicExpression[T]] = field(default=None)
    selected_variables: List[CanBehaveLikeAVariable[T]] = field(default_factory=list)
    warned_vars: typing.Set = field(default_factory=set, init=False)
    rule_mode: bool = field(default=False, init=False)

    def __post_init__(self):
        super().__post_init__()
        if in_symbolic_mode(EQLMode.Rule):
            self.rule_mode = True
        for variable in self.selected_variables:
            variable._var_._node_.enclosed = True

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(
        self, child: Optional[SymbolicExpression] = None, when_true: bool = True
    ):
        child = self._child_ if not child else child
        required_vars = self._parent_._required_variables_from_child_(
            self, when_true=when_true
        )
        required_vars.update(self.selected_variables)
        for var in self.selected_variables:
            required_vars.update(var._unique_variables_)
        if child and (when_true or (when_true is None)):
            for conc in child._conclusion_:
                required_vars.update(conc._unique_variables_)
        return required_vars

    def _evaluate_(
        self,
        selected_vars: Optional[Iterable[CanBehaveLikeAVariable]] = None,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        sources = sources or {}
        self._yield_when_false_ = yield_when_false
        if self._id_ in sources:
            self._is_false_ = self._id_expression_map_[self._id_]._is_false_
            yield sources
        self._inform_selected_variables_that_they_should_be_inferred_()
        if self._child_:
            self._child_._eval_parent_ = self
            child_values = self._child_._evaluate__(
                sources, yield_when_false=self._yield_when_false_
            )
        else:
            child_values = [{}]
        for v in child_values:
            v.update(sources)
            if self._child_:
                self._is_false_ = self._child_._is_false_
            if self._is_false_ and not self._yield_when_false_:
                continue
            if self._child_:
                for conclusion in self._child_._conclusion_:
                    v = conclusion._evaluate__(v)
            self._warn_on_unbound_variables_(v, selected_vars)
            if selected_vars:
                var_val_gen = {var: var._evaluate__(copy(v)) for var in selected_vars}
                original_v = v
                for sol in generate_combinations(var_val_gen):
                    v = copy(original_v)
                    var_val = {var._id_: sol[var][var._id_] for var in selected_vars}
                    v.update(var_val)
                    yield v
            else:
                yield v

    def _warn_on_unbound_variables_(
        self,
        sources: Dict[int, HashedValue],
        selected_vars: Iterable[CanBehaveLikeAVariable],
    ):
        """
        Warn the user if there are unbound variables in the query descriptor, because this will result in a cartesian
        product join operation.

        :param sources: The bound values after applying the conditions.
        :param selected_vars: The variables selected in the query descriptor.
        """
        unbound_variables = HashedIterable()
        for var in selected_vars:
            unbound_variables.update(
                var._unique_variables_.difference(HashedIterable(values=sources))
            )
        unbound_variables_with_domain = HashedIterable()
        for var in unbound_variables:
            if var.value._domain_ and len(var.value._domain_.values) > 20:
                if var not in self.warned_vars:
                    self.warned_vars.add(var)
                    unbound_variables_with_domain.add(var)
        if unbound_variables_with_domain:
            logger.warning(
                f"\nCartesian Product: "
                f"The following variables are not constrained "
                f"{unbound_variables_with_domain.unwrapped_values}"
                f"\nfor the query descriptor {self._name_}"
            )

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        vars = []
        if self.selected_variables:
            vars.extend(self.selected_variables)
        if self._child_:
            vars.extend(self._child_._all_variable_instances_)
        return vars

    def _inform_selected_variables_that_they_should_be_inferred_(self):
        if self.rule_mode and self._child_ and self._child_ is self._conditions_root_:
            for selected_variable in self.selected_variables:
                selected_variable._is_inferred_ = True

    def __repr__(self):
        return self._name_

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("ObjectDescriptor", "#d62728")

    @property
    def _name_(self) -> str:
        return f"({', '.join(var._name_ for var in self.selected_variables)})"


@dataclass(eq=False)
class SetOf(QueryObjectDescriptor[T]):
    """
    A query over a set of variables.
    """

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        self._yield_when_false_ = yield_when_false
        sol_gen = self._evaluate_(
            self.selected_variables, sources, yield_when_false=self._yield_when_false_
        )
        for sol in sol_gen:
            sol.update(sources)
            if self.selected_variables:
                var_val = {
                    var._id_: next(
                        var._evaluate__(sol, yield_when_false=self._yield_when_false_)
                    )[var._id_]
                    for var in self.selected_variables
                    if var._id_ in sol
                }
                sol.update(var_val)
                yield sol
            else:
                yield sol


@dataclass(eq=False)
class Entity(QueryObjectDescriptor[T]):
    """
    A query over a single variable.
    """

    def __post_init__(self):
        self._var_ = self.selected_variable
        super().__post_init__()

    @property
    def selected_variable(self):
        return self.selected_variables[0] if self.selected_variables else None

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[T]:
        self._yield_when_false_ = yield_when_false
        selected_variables = [self.selected_variable] if self.selected_variable else []
        sol_gen = self._evaluate_(
            selected_variables, sources, yield_when_false=self._yield_when_false_
        )
        for sol in sol_gen:
            sol.update(sources)
            if self._yield_when_false_ or not self._is_false_:
                if self.selected_variable:
                    for var_val in self.selected_variable._evaluate__(sol):
                        var_val.update(sol)
                        yield var_val
                else:
                    yield sol


@dataclass(eq=False)
class Infer(An[T]):

    def __post_init__(self):
        super().__post_init__()
        for v in self._child_.selected_variables:
            v._is_inferred_ = True
        self._node_.wrap_subtree = False

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("Infer", "#EAC9FF")


@dataclass
class From:
    """
    A dataclass that holds the domain for a symbolic variable, this will be used instead of the global cache
    of the variable class type.
    """

    domain: Any
    """
    The domain to use for the symbolic variable.
    """


@dataclass(eq=False)
class Variable(CanBehaveLikeAVariable[T]):
    _name__: str
    """
    The name of the variable.
    """
    _type_: Type
    """
    The class that this variable represents.
    """
    _kwargs_: Dict[str, Any] = field(default_factory=dict)
    """
    The properties of the variable as keyword arguments.
    """
    _domain_source_: Optional[From] = field(default=None, kw_only=True, repr=False)
    """
    An optional source for the variable domain. If not given, the global cache of the variable class type will be used
    as the domain, or if kwargs are given the type and the kwargs will be used to create/infer new values for the
    variable.
    """
    _domain_: HashedIterable = field(
        default_factory=HashedIterable, kw_only=True, repr=False
    )
    """
    The iterable domain of values for this variable.
    """
    _invert_: bool = field(init=False, default=False, repr=False)
    """
    Redefined from super class to give it a default value.
    """
    _predicate_type_: Optional[PredicateType] = field(default=None, repr=False)
    """
    If this symbol is an instance of the Predicate class.
    """
    _is_inferred_: bool = field(default=False, repr=False)
    """
    Whether this variable should be inferred.
    """
    _is_indexed_: bool = field(default=True, repr=False)
    """
    Whether this variable cache is indexed or flat.
    """
    _cache_: ClassVar[Dict[Type, IndexedCache]] = defaultdict(IndexedCache)
    """
    A mapping from variable type to an indexed cache of all seen inputs and outputs of the variable type. 
    """
    _child_vars_: Optional[Dict[str, SymbolicExpression]] = field(
        default_factory=dict, init=False, repr=False
    )
    """
    A dictionary mapping child variable names to variables, these are from the _kwargs_ dictionary. 
    """
    _kwargs_expression_: Optional[SymbolicExpression] = field(
        default=None, init=False, repr=False
    )
    """
    An expression of the constraints added from the keyword arguments of the variable.
    """
    _evaluating_kwargs_expression_: bool = field(default=False, init=False, repr=False)
    """
    A flag indicating that the kwargs expression is currently being evaluated so do not evaluate them again, and instead
    yield from the domain.
    """

    def __post_init__(self):
        self._validate_inputs_and_fill_missing_ones_()
        self._var_ = self
        super().__post_init__()
        # has to be after super init because this needs the node of this variable to be initialized first.
        self._update_child_vars_from_kwargs_()

    def _validate_inputs_and_fill_missing_ones_(self):
        if self._kwargs_ and not self._type_:
            raise ValueError(
                f"Variable {self._name_} has class keyword arguments but no type is specified."
            )
        self._child_ = None
        if self._domain_source_:
            self._update_domain_(self._domain_source_.domain)

    def _update_domain_(self, domain):
        if domain:
            new_domain = None
            if isinstance(domain, HashedIterable):
                self._domain_ = domain
            if isinstance(domain, SymbolicExpression):
                new_domain = (v[domain._id_] for v in domain._evaluate__())
            elif not is_iterable(domain):
                new_domain = [HashedValue(domain)]
            new_domain = new_domain or domain
            self._domain_.set_iterable(new_domain)

    def _update_child_vars_from_kwargs_(self):
        if self._kwargs_:
            for k, v in self._kwargs_.items():
                if isinstance(v, SymbolicExpression):
                    self._child_vars_[k] = v
                else:
                    self._child_vars_[k] = Literal(v, name=k)
            self._update_children_(*self._child_vars_.values())

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        """
        A variable either is already bound in sources by other constraints (Symbolic Expressions).,
        or has no domain and will instantiate new values by constructing the type if the type is given,
        or will yield from current domain if exists.
        """
        self._yield_when_false_ = yield_when_false
        sources = sources or {}
        if self._id_ in sources:
            if self is self._conditions_root_ or isinstance(
                self._parent_, LogicalOperator
            ):
                original_me = self._id_expression_map_[self._id_]
                self._is_false_ = original_me._is_false_
                if not original_me._is_false_ or self._yield_when_false_:
                    yield sources
            else:
                yield sources
        elif self._domain_ and not self._is_inferred_:
            if self._kwargs_expression_ and not self._evaluating_kwargs_expression_:
                # because when kwargs expression exists,
                # it will constrain the domain further to fit the kwargs provided.
                yield from self._evaluate_kwargs_expression_(sources)
            else:
                # If no kwargs expression, or is currently being evaluated then yield from the domain directly,
                # if ht kwargs is being evaluated, it will want to take the domain from here and constrain it further.
                yield from self
        elif not self._is_inferred_ and not self._predicate_type_:
            self._update_domain_and_kwargs_expression_()
            yield from self._evaluate__(
                sources, yield_when_false=self._yield_when_false_
            )
        elif self._child_vars_:
            for kwargs in self._generate_combinations_for_child_vars_values_(sources):
                yield from self._yield_from_cache_or_instantiate_new_values_(
                    sources, kwargs
                )

    def _evaluate_kwargs_expression_(
        self, sources: Optional[Dict[int, HashedValue]] = None
    ):
        self._evaluating_kwargs_expression_ = True
        for v in self._kwargs_expression_._evaluate__(
            sources, yield_when_false=self._yield_when_false_
        ):
            if self is self._conditions_root_ or isinstance(
                self._parent_, LogicalOperator
            ):
                self._is_false_ = self._kwargs_expression_._is_false_
                if not self._is_false_ or self._yield_when_false_:
                    yield v
            else:
                yield v
        self._evaluating_kwargs_expression_ = False

    def _update_domain_and_kwargs_expression_(self):
        self._domain_source_ = From(self._cache_values_)
        self._update_domain_(self._domain_source_.domain)
        if self._kwargs_:
            parents = [p for p in self._node_.parents]
            self._kwargs_expression_, attributes = properties_to_expression_tree(
                self, self._child_vars_
            )
            self._kwargs_expression_ = An(Entity(self._kwargs_expression_, [self]))
            self._replace_expression_with_(self._kwargs_expression_, parents)

    def _replace_expression_with_(
        self, new_expression: SymbolicExpression, parents: Optional[List] = None
    ):
        if not parents:
            parents = [
                p
                for p in self._node_.parents
                if new_expression._node_ not in p.ancestors
            ]
        for child in self._node_.children:
            self._node_.remove_child(child)
        for parent in parents:
            new_expression._parent_ = parent.data
            self._node_.remove_parent(parent)
        self._node_.remove()
        self._node_ = new_expression._node_

    def _instantiate_new_values_or_yield_from_cache_(
        self, sources: Optional[Dict[int, HashedValue]] = None
    ) -> Iterable[Dict[int, HashedValue]]:
        # Precompute generators only when we have child vars
        if self._child_vars_:
            for kwargs in self._generate_combinations_for_child_vars_values_(sources):
                yield from self._yield_from_cache_or_instantiate_new_values_(
                    sources, kwargs
                )
        else:
            yield from self._yield_from_cache_or_instantiate_new_values_(sources)

    def _generate_combinations_for_child_vars_values_(
        self, sources: Optional[Dict[int, HashedValue]] = None
    ):
        kwargs_generators = {
            k: v._evaluate__(sources) for k, v in self._child_vars_.items()
        }
        yield from generate_combinations(kwargs_generators)

    def _yield_from_cache_or_instantiate_new_values_(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        kwargs: Dict[str, Dict[int, HashedValue]] = None,
    ):
        kwargs = kwargs or {}
        # Try cache fast-path first when allowed
        retrieved = False
        if not self._is_inferred_ and self._is_indexed_:
            for v in self._search_and_yield_from_cache_(kwargs):
                retrieved = True
                if sources:
                    v.update(sources)
                yield v

        # If nothing retrieved and we are allowed to instantiate
        if (not retrieved) and (self._is_inferred_ or self._predicate_type_):
            yield from self._instantiate_new_values_and_yield_results_(kwargs, sources)

    def _instantiate_new_values_and_yield_results_(
        self,
        kwargs: Dict[str, Dict[int, HashedValue]],
        sources: Optional[Dict[int, HashedValue]] = None,
    ) -> Iterable[Dict[int, HashedValue]]:
        # Build once: unwrapped hashed kwargs for already provided child vars
        bound_kwargs = {k: v[self._child_vars_[k]._id_] for k, v in kwargs.items()}
        # For missing kwargs, evaluate their generators lazily
        unbound_kwargs = {
            k: v._evaluate__(sources)
            for k, v in self._child_vars_.items()
            if k not in bound_kwargs
        }
        if unbound_kwargs:
            yield from self._bind_unbound_kwargs_and_yield_results_(
                kwargs, unbound_kwargs, bound_kwargs
            )
        else:
            instance = self._type_(**{k: hv.value for k, hv in bound_kwargs.items()})
            yield from self._process_output_and_update_values_(instance, **kwargs)

    def _bind_unbound_kwargs_and_yield_results_(
        self,
        kwargs: Dict[str, Dict[int, HashedValue]],
        unbound_kwargs: Dict[str, Iterable],
        bound_kwargs: Dict[str, HashedValue],
    ):
        for extra_kwargs in generate_combinations(unbound_kwargs):
            # Avoid mutating the shared kwargs dict; work on a shallow copy
            merged_kwargs = dict(kwargs)
            merged_kwargs.update(extra_kwargs)
            # Update unwrapped hashed args from the delta only
            for k, v in extra_kwargs.items():
                bound_kwargs[k] = v[self._child_vars_[k]._id_]
            instance = self._type_(**{k: hv.value for k, hv in bound_kwargs.items()})
            yield from self._process_output_and_update_values_(
                instance, **merged_kwargs
            )

    def _search_and_yield_from_cache_(self, kwargs: Optional[Dict] = None):
        unwrapped_hashed_kwargs = None
        if kwargs and self._is_indexed_:
            unwrapped_hashed_kwargs = {
                k: v[self._child_vars_[k]._id_] for k, v in kwargs.items()
            }
        for _, value in yield_class_values_from_cache(
            self._cache_,
            self._type_,
            unwrapped_hashed_kwargs,
            from_index=self._is_indexed_,
        ):
            yield from self._process_output_and_update_values_(value.value, **kwargs)

    @property
    def _cache_values_(self):
        for _, value in yield_class_values_from_cache(
            self._cache_, self._type_, from_index=self._is_indexed_
        ):
            yield value

    @property
    def _cache_keys_(self) -> List[Type]:
        """
        Get the cache keys for the given class which are its subclasses and itself.
        """
        return get_cache_keys_for_class_(self._cache_, self._type_)

    def _process_output_and_update_values_(
        self, function_output: Any, **kwargs
    ) -> Iterable[Dict[int, HashedValue]]:
        """
        Process the output of the predicate/variable and get the results.

        :param function_output: The output of the predicate.
        :param kwargs: The keyword arguments of the predicate/variable.
        :return: The results' dictionary.
        """
        # evaluate the predicate.
        if self._predicate_type_ == PredicateType.SubClassOfPredicate:
            function_output = function_output()

        # Compute truth considering inversion
        result_truthy = bool(function_output)
        self._is_false_ = result_truthy if self._invert_ else not result_truthy

        if self._yield_when_false_ or not self._is_false_:
            hv = (
                function_output
                if isinstance(function_output, HashedValue)
                else HashedValue(function_output)
            )

            if not kwargs:
                yield {self._id_: hv}
                return

            # kwargs is a mapping from name -> {var_id: HashedValue};
            # we need a single dict {var_id: HashedValue
            values = {self._id_: hv}
            for d in kwargs.values():
                values.update(d)
            yield values

    @property
    def _name_(self):
        return self._name__

    @classmethod
    def _from_domain_(
        cls, iterable, clazz: Optional[Type] = None, name: Optional[str] = None
    ) -> Variable:
        if not isinstance(iterable, SymbolicExpression) and not is_iterable(iterable):
            iterable = HashedIterable([iterable])
        if not clazz:
            clazz = type(next(iter(iterable)))
        if name is None:
            name = clazz.__name__
        return Variable(name, clazz, _domain_source_=From(iterable))

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        variables = [self]
        for v in self._child_vars_.values():
            variables.extend(v._all_variable_instances_)
        return variables

    @property
    def _plot_color_(self) -> ColorLegend:
        if self._plot_color__:
            return self._plot_color__
        else:
            return ColorLegend("Variable", "cornflowerblue")

    @_plot_color_.setter
    def _plot_color_(self, value: ColorLegend):
        self._plot_color__ = value
        self._node_.color = value

    def __iter__(self):
        for v in self._domain_:
            yield {self._id_: HashedValue(v)}

    def __repr__(self):
        return self._name_


@dataclass(eq=False)
class Literal(Variable[T]):
    """
    Literals are variables that are not constructed by their type but by their given data.
    """

    def __init__(
        self, data: Any, name: Optional[str] = None, type_: Optional[Type] = None
    ):
        original_data = data
        data = [data]
        if not is_iterable(data):
            data = HashedIterable([data])
        if not type_:
            first_value = next(iter(data), None)
            type_ = type(first_value) if first_value else None
        if name is None:
            if type_:
                name = type_.__name__
            else:
                name = type(original_data).__name__
        super().__init__(name, type_, _domain_source_=From(data))

    @property
    def _plot_color_(self) -> ColorLegend:
        if self._plot_color__:
            return self._plot_color__
        else:
            return ColorLegend("Literal", "#949292")


@dataclass(eq=False)
class DomainMapping(CanBehaveLikeAVariable[T], ABC):
    """
    A symbolic expression the maps the domain of symbolic variables.
    """

    _child_: CanBehaveLikeAVariable[T]
    _invert_: bool = field(init=False, default=False)

    def __post_init__(self):
        super().__post_init__()
        self._var_ = self

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        return self._child_._all_variable_instances_

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        sources = sources or {}
        self._yield_when_false_ = yield_when_false
        self._child_._eval_parent_ = self
        if self._id_ in sources:
            yield sources
            return
        child_val = self._child_._evaluate__(
            sources, yield_when_false=self._yield_when_false_
        )
        for child_v in child_val:
            for v in self._apply_mapping_(child_v[self._child_._id_]):
                values = copy(child_v)
                if (not self._invert_ and v.value) or (self._invert_ and not v.value):
                    self._is_false_ = False
                else:
                    self._is_false_ = True
                if self._yield_when_false_ or not self._is_false_:
                    values[self._id_] = v
                    yield values

    @abstractmethod
    def _apply_mapping_(self, value: HashedValue) -> Iterable[HashedValue]:
        """
        Apply the domain mapping to a symbolic value.
        """
        pass

    @property
    def _plot_color_(self) -> ColorLegend:
        if self._plot_color__:
            return self._plot_color__
        else:
            return ColorLegend("DomainMapping", "#8FC7B8")

    @_plot_color_.setter
    def _plot_color_(self, value: ColorLegend):
        self._plot_color__ = value
        self._node_.color = value


@dataclass(eq=False)
class Attribute(DomainMapping):
    """
    A symbolic attribute that can be used to access attributes of symbolic variables.
    """

    _attr_name_: str

    def _apply_mapping_(self, value: HashedValue) -> Iterable[HashedValue]:
        yield HashedValue(id_=value.id_, value=getattr(value.value, self._attr_name_))

    @property
    def _name_(self):
        return f"{self._child_._var_._name_}.{self._attr_name_}"


@dataclass(eq=False)
class Index(DomainMapping):
    """
    A symbolic indexing operation that can be used to access items of symbolic variables via [] operator.
    """

    _key_: Any

    def _apply_mapping_(self, value: HashedValue) -> Iterable[HashedValue]:
        yield HashedValue(id_=value.id_, value=value.value[self._key_])

    @property
    def _name_(self):
        return f"{self._child_._var_._name_}[{self._key_}]"


@dataclass(eq=False)
class Call(DomainMapping):
    """
    A symbolic call that can be used to call methods on symbolic variables.
    """

    _args_: Tuple[Any, ...] = field(default_factory=tuple)
    _kwargs_: Dict[str, Any] = field(default_factory=dict)

    def _apply_mapping_(self, value: HashedValue) -> Iterable[HashedValue]:
        if len(self._args_) > 0 or len(self._kwargs_) > 0:
            yield HashedValue(
                id_=value.id_, value=value.value(*self._args_, **self._kwargs_)
            )
        else:
            yield HashedValue(id_=value.id_, value=value.value())

    @property
    def _name_(self):
        return f"{self._child_._var_._name_}()"


@dataclass(eq=False)
class Flatten(DomainMapping):
    """
    Domain mapping that flattens an iterable-of-iterables into a single iterable of items.

    Given a child expression that evaluates to an iterable (e.g., Views.bodies), this mapping yields
    one solution per inner element while preserving the original bindings (e.g., the View instance),
    similar to UNNEST in SQL.
    """

    def __post_init__(self):
        if not isinstance(self._child_, SymbolicExpression):
            self._child_ = Literal(self._child_)
        super().__post_init__()

    def _apply_mapping_(self, value: HashedValue) -> Iterable[HashedValue]:
        inner = value.value
        # Treat non-iterables as singletons
        if not is_iterable(inner):
            inner_iter = [inner]
        else:
            inner_iter = inner
        for inner_v in inner_iter:
            yield HashedValue(inner_v)

    @property
    def _name_(self):
        return f"Flatten({self._child_._name_})"


@dataclass(eq=False)
class Concatenate(CanBehaveLikeAVariable[T]):
    _child_: CanBehaveLikeAVariable[T]
    _invert_: bool = field(init=False, default=False)

    def __post_init__(self):
        super().__post_init__()
        self._var_ = self

    def _evaluate__(
        self, sources: Optional[Dict[int, HashedValue]] = None
    ) -> Iterable[Dict[int, HashedValue]]:
        sources = sources or {}
        if self._id_ in sources:
            yield sources
            return
        all_values = defaultdict(list)
        for child_v in self._child_._evaluate__(sources):
            child_v = copy(child_v)
            for id_, val in child_v.items():
                if id_ == self._child_._id_:
                    child_v_unwrapped = val.value
                    if not is_iterable(child_v_unwrapped):
                        child_v_unwrapped = [child_v_unwrapped]
                    all_values[self._id_].extend(child_v_unwrapped)
                all_values[id_].append(val)
            for s_id, s_val in sources.items():
                all_values[s_id].append(s_val)
        yield {k: HashedValue(v) for k, v in all_values.items()}

    @property
    def _name_(self):
        return f"{self.__class__.__name__}({self._child_._name_})"

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        return self._child_._all_variable_instances_


@dataclass(eq=False)
class BinaryOperator(SymbolicExpression, ABC):
    """
    A base class for binary operators that can be used to combine symbolic expressions.
    """

    left: SymbolicExpression
    right: SymbolicExpression
    _child_: SymbolicExpression = field(init=False, default=None)
    _cache_: IndexedCache = field(default_factory=IndexedCache, init=False)

    def __post_init__(self):
        super().__post_init__()
        self.left, self.right = self._update_children_(self.left, self.right)
        combined_vars = self.left._unique_variables_.union(
            self.right._unique_variables_
        )
        self._cache_.keys = [
            v.id_
            for v in combined_vars.filter(lambda v: not isinstance(v.value, Literal))
        ]

    def yield_final_output_from_cache(
        self, variables_sources, cache: Optional[IndexedCache] = None
    ) -> Iterable[Dict[int, HashedValue]]:
        cache = self._cache_ if cache is None else cache
        entered = False
        for output, is_false in cache.retrieve(variables_sources):
            entered = True
            self._is_false_ = is_false
            cache_match_count.values[self._node_.name] += 1
            if is_false and self._is_duplicate_output_(output):
                continue
            yield output
        if not entered:
            cache_match_count.values[self._node_.name] += 1
        cache_enter_count.values[self._node_.name] = cache.enter_count
        cache_search_count.values[self._node_.name] = cache.search_count

    def yield_from_cache(
        self, variables_sources, cache: IndexedCache
    ) -> Iterable[Tuple[Dict[int, HashedValue], bool]]:
        entered = False
        for output, is_false in cache.retrieve(variables_sources):
            entered = True
            cache_match_count.values[self._node_.name] += 1
            yield output, is_false
        if not entered:
            cache_match_count.values[self._node_.name] += 1
        cache_enter_count.values[self._node_.name] = cache.enter_count
        cache_search_count.values[self._node_.name] = cache.search_count

    def update_cache(
        self, values: Dict[int, HashedValue], cache: Optional[IndexedCache] = None
    ):
        if not is_caching_enabled():
            return
        cache = self._cache_ if cache is None else cache
        cache.insert(
            {k: v for k, v in values.items() if k in cache.keys}, output=self._is_false_
        )

    @property
    @lru_cache(maxsize=None)
    def _all_variable_instances_(self) -> List[Variable]:
        """
        Get the leaf instances of the symbolic expression.
        This is useful for accessing the leaves of the symbolic expression tree.
        """
        return self.left._all_variable_instances_ + self.right._all_variable_instances_

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(
        self, child: Optional[SymbolicExpression] = None, when_true: bool = True
    ):
        if not child:
            child = self.left
        required_vars = HashedIterable()
        if child is self.left:
            required_vars.update(self.right._unique_variables_)
        if when_true or (when_true is None):
            for conc in self._conclusion_:
                required_vars.update(conc._unique_variables_)
        if self._parent_:
            required_vars.update(
                self._parent_._required_variables_from_child_(self, when_true)
            )
        return required_vars


@dataclass(eq=False)
class ForAll(BinaryOperator):

    solution_set: List[Dict[int, HashedValue]] = field(init=False, default_factory=list)

    @property
    def _name_(self) -> str:
        return self.__class__.__name__

    @property
    def variable(self):
        return self.left

    @variable.setter
    def variable(self, value):
        self.left = value

    @property
    def condition(self):
        return self.right

    @condition.setter
    def condition(self, value):
        self.right = value

    @property
    @lru_cache(maxsize=None)
    def condition_unique_variable_ids(self) -> List[int]:
        return [
            v.id_
            for v in self.condition._unique_variables_.difference(
                self.left._unique_variables_
            )
        ]

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        sources = sources or {}

        # Always reset per evaluation
        self.solution_set = []

        var_val_index = 0

        for var_val in self.variable._evaluate__(sources):
            ctx = {**sources, **var_val}
            current = []

            # Evaluate the condition under this particular universal value
            for condition_val in self.condition._evaluate__(ctx):
                if self.condition._is_false_:
                    continue
                # Keep only the non-universal variables from the condition bindings
                filtered = {
                    k: v
                    for k, v in condition_val.items()
                    if k in self.condition_unique_variable_ids
                }
                current.append(filtered)

            # If the condition yields no satisfying bindings for this universal value, the universal fails
            if not current:
                self.solution_set = []
                break

            if var_val_index == 0:
                # seed with all satisfying non-universal bindings
                self.solution_set = current
            else:
                # Intersect with previously accumulated satisfying bindings
                current_set = {tuple(sorted(d.items())) for d in current}
                self.solution_set = [
                    d
                    for d in self.solution_set
                    if tuple(sorted(d.items())) in current_set
                ]

            var_val_index += 1

            # Early exit if the intersection is empty
            if not self.solution_set:
                break

        # Yield the remaining bindings (non-universal) merged with the incoming sources
        for sol in self.solution_set or []:
            out = copy(sol)
            out.update(sources)
            yield out


@dataclass(eq=False)
class Comparator(BinaryOperator):
    """
    A symbolic equality check that can be used to compare symbolic variables.
    """

    left: CanBehaveLikeAVariable
    right: CanBehaveLikeAVariable
    operation: Callable[[Any, Any], bool]
    _invert__: bool = field(init=False, default=False)
    operation_name_map: ClassVar[Dict[Any, str]] = {
        operator.eq: "==",
        operator.ne: "!=",
        operator.lt: "<",
        operator.le: "<=",
        operator.gt: ">",
        operator.ge: ">=",
    }

    @property
    def _invert_(self):
        return self._invert__

    @_invert_.setter
    def _invert_(self, value):
        if value == self._invert__:
            return
        self._invert__ = value
        prev_operation = self.operation
        match self.operation:
            case operator.lt:
                self.operation = operator.ge if self._invert_ else self.operation
            case operator.gt:
                self.operation = operator.le if self._invert_ else self.operation
            case operator.le:
                self.operation = operator.gt if self._invert_ else self.operation
            case operator.ge:
                self.operation = operator.lt if self._invert_ else self.operation
            case operator.eq:
                self.operation = operator.ne if self._invert_ else self.operation
            case operator.ne:
                self.operation = operator.eq if self._invert_ else self.operation
            case operator.contains:

                def not_contains(a, b):
                    return not operator.contains(a, b)

                self.operation = not_contains if self._invert_ else self.operation
            case _:
                raise ValueError(f"Unsupported operation: {self.operation.__name__}")
        self._node_.name = self._node_.name.replace(
            prev_operation.__name__, self.operation.__name__
        )

    @property
    def _name_(self):
        if self.operation in self.operation_name_map:
            return self.operation_name_map[self.operation]
        return self.operation.__name__

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        """
        Compares the left and right symbolic variables using the "operation".

        :param sources: Dictionary of symbolic variable id to a value of that variable, the left and right values
        will retrieve values from sources if they exist, otherwise will directly retrieve them from the original
        sources.
        :return: Yields a HashedIterable mapping a symbolic variable id to a value of that variable, it will contain
         only two values, the left and right symbolic values.
        """
        sources = sources or {}
        self._yield_when_false_ = yield_when_false

        if self._id_ in sources:
            yield sources
            return

        if is_caching_enabled():
            if self._cache_.check(sources):
                yield from self.yield_final_output_from_cache(sources)
                return

        first_operand, second_operand = self.get_first_second_operands(sources)
        first_operand._eval_parent_ = self
        first_values = first_operand._evaluate__(sources)
        for first_value in first_values:
            first_value.update(sources)
            operand_value_map = {first_operand._id_: first_value[first_operand._id_]}
            second_operand._eval_parent_ = self
            second_values = second_operand._evaluate__(first_value)
            for second_value in second_values:
                operand_value_map[second_operand._id_] = second_value[
                    second_operand._id_
                ]
                res = self.apply_operation(operand_value_map)
                self._is_false_ = not res
                if res or self._yield_when_false_:
                    values = copy(first_value)
                    values.update(second_value)
                    values.update(operand_value_map)
                    values[self._id_] = HashedValue(res)
                    self.update_cache(values)
                    yield values

    def apply_operation(self, operand_values: Dict[int, HashedValue]):
        return self.operation(
            operand_values[self.left._id_].value, operand_values[self.right._id_].value
        )

    def get_first_second_operands(
        self, sources: Dict[int, HashedValue]
    ) -> Tuple[SymbolicExpression, SymbolicExpression]:
        if sources and any(
            v.value._var_._id_ in sources for v in self.right._unique_variables_
        ):
            return self.right, self.left
        else:
            return self.left, self.right

    def get_result_domain(
        self, operand_value_map: Dict[CanBehaveLikeAVariable, HashedValue]
    ) -> HashedIterable:
        left_leaf_value = self.left._var_._domain_[operand_value_map[self.left].id_]
        right_leaf_value = self.right._var_._domain_[operand_value_map[self.right].id_]
        return HashedIterable(
            values={
                self.left._var_._id_: left_leaf_value,
                self.right._var_._id_: right_leaf_value,
            }
        )

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("Comparator", "#ff7f0e")


@dataclass(eq=False)
class LogicalOperator(BinaryOperator, ABC):
    """
    A symbolic operation that can be used to combine multiple symbolic expressions.
    """

    right_cache: IndexedCache = field(default_factory=IndexedCache, init=False)

    def __post_init__(self):
        super().__post_init__()
        right_vars = self.right._unique_variables_.filter(
            lambda v: not isinstance(v, Literal)
        )
        self.right_cache.keys = [v.id_ for v in right_vars]

    @property
    def _name_(self):
        return self.__class__.__name__

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("LogicalOperator", "#2ca02c")


@dataclass(eq=False, repr=False)
class AND(LogicalOperator):
    """
    A symbolic AND operation that can be used to combine multiple symbolic expressions.
    """

    seen_left_values: SeenSet = field(default_factory=SeenSet, init=False)

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        # init an empty source if none is provided
        sources = sources or {}
        self._yield_when_false_ = yield_when_false

        # constrain left values by available sources
        left_prev = self.left._eval_parent_
        self.left._eval_parent_ = self
        try:
            left_values = self.left._evaluate__(
                sources, yield_when_false=self._yield_when_false_
            )
            for left_value in left_values:
                left_value.update(sources)
                if self._yield_when_false_ and self.left._is_false_:
                    self._is_false_ = True
                    if self._is_duplicate_output_(left_value):
                        continue
                    yield left_value
                    continue

                if is_caching_enabled() and self.right_cache.check(left_value):
                    yield from self.yield_final_output_from_cache(
                        left_value, self.right_cache
                    )
                    continue

                # constrain right values by available sources
                right_prev = self.right._eval_parent_
                self.right._eval_parent_ = self
                try:
                    right_values = self.right._evaluate__(
                        left_value, yield_when_false=self._yield_when_false_
                    )

                    # For the found left value, find all right values,
                    # and yield the (left, right) results found.
                    for right_value in right_values:
                        output = copy(right_value)
                        output.update(left_value)
                        self._is_false_ = self.right._is_false_
                        self.update_cache(right_value, self.right_cache)
                        yield output
                finally:
                    self.right._eval_parent_ = right_prev
        finally:
            self.left._eval_parent_ = left_prev


@dataclass(eq=False)
class OR(LogicalOperator, ABC):
    """
    A symbolic single choice operation that can be used to choose between multiple symbolic expressions.
    """

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(
        self,
        child: Optional[SymbolicExpression] = None,
        when_true: Optional[bool] = None,
    ):
        if not child:
            child = self.left
        required_vars = HashedIterable()
        when_false = not when_true if when_true is not None else None
        if child is self.left:
            if when_false or (when_false is None):
                required_vars.update(self.right._unique_variables_)
                when_iam = None
            else:
                when_iam = True
            if self._parent_:
                required_vars.update(
                    self._parent_._required_variables_from_child_(self, when_iam)
                )
            if when_true or (when_true is None):
                for conc in self.left._conclusion_:
                    required_vars.update(conc._unique_variables_)
        elif child is self.right:
            if when_true or (when_true is None):
                for conc in self.right._conclusion_:
                    required_vars.update(conc._unique_variables_)
            if self._parent_:
                required_vars.update(
                    self._parent_._required_variables_from_child_(self, when_true)
                )
        return required_vars


@dataclass(eq=False)
class Union(OR):
    left_cache: IndexedCache = field(default_factory=IndexedCache, init=False)
    left_evaluated: bool = field(default=False, init=False)
    right_evaluated: bool = field(default=False, init=False)

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        # init an empty source if none is provided
        sources = sources or {}
        self._yield_when_false_ = yield_when_false

        if is_caching_enabled() and self._cache_.check(sources):
            yield from self.yield_final_output_from_cache(sources)
            return

        # constrain left values by available sources
        left_prev = self.left._eval_parent_
        self.left._eval_parent_ = self
        try:
            left_values = self.left._evaluate__(
                sources, yield_when_false=self._yield_when_false_
            )

            for left_value in left_values:
                output = copy(sources)
                output.update(left_value)
                self.left_evaluated = True
                if self.left._is_false_:
                    if self._yield_when_false_:
                        yield from self.evaluate_right(output)
                    continue
                if self._is_duplicate_output_(output):
                    continue
                self.update_cache(output, self._cache_)
                yield output
        finally:
            self.left._eval_parent_ = left_prev
        self.left_evaluated = False
        yield from self.evaluate_right(sources)

    def evaluate_right(
        self, sources: Optional[Dict[int, HashedValue]]
    ) -> Iterable[Dict[int, HashedValue]]:
        right_values = self.right._evaluate__(
            sources, yield_when_false=self._yield_when_false_
        )
        # For the found left value, find all right values,
        # and yield the (left, right) results found.
        for right_value in right_values:
            sources.update(right_value)
            if self._yield_when_false_ and self.left_evaluated:
                self._is_false_ = self.left._is_false_ and self.right._is_false_
            else:
                self._is_false_ = False
            if not self._is_false_:
                if self._is_duplicate_output_(sources):
                    continue
            self.right_evaluated = True
            self.update_cache(sources, self._cache_)
            yield sources


@dataclass(eq=False)
class ElseIf(OR):
    """
    A symbolic single choice operation that can be used to choose between multiple symbolic expressions.
    """

    def _evaluate__(
        self,
        sources: Optional[Dict[int, HashedValue]] = None,
        yield_when_false: bool = False,
    ) -> Iterable[Dict[int, HashedValue]]:
        """
        Constrain the symbolic expression based on the indices of the operands.
        This method overrides the base class method to handle ElseIf logic.
        """
        # init an empty source if none is provided
        sources = sources or {}
        self._yield_when_false_ = yield_when_false

        # constrain left values by available sources
        left_prev = self.left._eval_parent_
        self.left._eval_parent_ = self
        try:
            # Force yield_when_false=True for the left branch to preserve else-if semantics
            left_values = self.left._evaluate__(sources, yield_when_false=True)
            any_left = False
            for left_value in left_values:
                any_left = True
                left_value.update(sources)
                if self.left._is_false_:
                    if is_caching_enabled() and self.right_cache.check(left_value):
                        yield from self.yield_final_output_from_cache(
                            left_value, self.right_cache
                        )
                        continue
                    right_prev = self.right._eval_parent_
                    self.right._eval_parent_ = self
                    try:
                        right_values = self.right._evaluate__(
                            left_value, yield_when_false=self._yield_when_false_
                        )
                        for right_value in right_values:
                            self._is_false_ = self.right._is_false_
                            output = copy(left_value)
                            output.update(right_value)
                            if self._is_false_ and not self._yield_when_false_:
                                continue
                            if not self._is_false_:
                                if self._is_duplicate_output_(output):
                                    continue
                            self.update_cache(right_value, self.right_cache)
                            yield output
                    finally:
                        self.right._eval_parent_ = right_prev
                else:
                    self._is_false_ = False
                    yield left_value
            # If left produced no values at all, evaluate right against sources
            if not any_left:
                right_prev = self.right._eval_parent_
                self.right._eval_parent_ = self
                try:
                    right_values = self.right._evaluate__(
                        sources, yield_when_false=self._yield_when_false_
                    )
                    for right_value in right_values:
                        self._is_false_ = self.right._is_false_
                        if self._is_false_ and not self._yield_when_false_:
                            continue
                        self.update_cache(right_value, self.right_cache)
                        yield right_value
                finally:
                    self.right._eval_parent_ = right_prev
        finally:
            self.left._eval_parent_ = left_prev


def Not(operand: Any) -> SymbolicExpression:
    """
    A symbolic NOT operation that can be used to negate symbolic expressions.
    """
    if not isinstance(operand, SymbolicExpression):
        operand = Literal(operand)
    if isinstance(operand, ResultQuantifier):
        raise NotImplementedError(
            f"Symbolic NOT operations on {ResultQuantifier} operands "
            f"are not allowed, you can negate the conditions or {QueryObjectDescriptor}"
            f" instead as negating quantifiers is most likely not what you want"
            f" as it is ambiguous."
        )
    elif isinstance(operand, Entity):
        operand = operand.__class__(Not(operand._child_), operand.selected_variables)
    elif isinstance(operand, SetOf):
        operand = operand.__class__(Not(operand._child_), operand.selected_variables)
    elif isinstance(operand, AND):
        operand = ElseIf(Not(operand.left), Not(operand.right))
    elif isinstance(operand, OR):
        operand = AND(Not(operand.left), Not(operand.right))
    else:
        operand._invert_ = True
    return operand


OperatorOptimizer = Callable[[SymbolicExpression, SymbolicExpression], LogicalOperator]


def chained_logic(
    operator: TypingUnion[Type[LogicalOperator], OperatorOptimizer], *conditions
):
    """
    A chian of logic operation over multiple conditions, e.g. cond1 | cond2 | cond3.

    :param operator: The symbolic operator to apply between the conditions.
    :param conditions: The conditions to be chained.
    """
    prev_operation = None
    for condition in conditions:
        if prev_operation is None:
            prev_operation = condition
            continue
        prev_operation = operator(prev_operation, condition)
    return prev_operation


@contextmanager
def rule_mode(query: Optional[SymbolicExpression] = None):
    """
    Wrapper around symbolic construction mode to easily enable rule mode
    """
    # delegate to symbolic_mode
    with symbolic_mode(query, EQLMode.Rule) as ctx:
        yield ctx


@contextmanager
def symbolic_mode(
    query: Optional[SymbolicExpression] = None, mode: EQLMode = EQLMode.Query
):
    """
    Context manager to temporarily enable symbolic construction mode.

    Within the context, calling classes decorated with ``@symbol`` produces
    symbolic Variables instead of real instances.

    :param query: Optional symbolic expression to also enter/exit as a context.
    """
    prev_mode = _symbolic_mode.get()
    try:
        if query is not None:
            query.__enter__(in_rule_mode=True)
        _set_symbolic_mode(mode)
        yield SymbolicExpression._current_parent_()
    finally:
        if query is not None:
            query.__exit__()
        _set_symbolic_mode(prev_mode)


def properties_to_expression_tree(
    var: CanBehaveLikeAVariable, properties: Dict[str, Any]
) -> Tuple[SymbolicExpression, List[SymbolicExpression]]:
    """
    Convert properties of a variable to a symbolic expression.
    """
    with symbolic_mode():
        conditions = [getattr(var, k) == v for k, v in properties.items()]
        expression = None
        if len(conditions) == 1:
            expression = conditions[0]
        elif len(conditions) > 1:
            expression = chained_logic(AND, *conditions)
    return expression, [op.left for op in conditions]


def _optimize_or(left: SymbolicExpression, right: SymbolicExpression) -> OR:
    with symbolic_mode(mode=None):
        left_vars = left._unique_variables_.filter(
            lambda v: not isinstance(v.value, Literal)
        )
        right_vars = right._unique_variables_.filter(
            lambda v: not isinstance(v.value, Literal)
        )
        if set(left_vars.unwrapped_values) == set(right_vars.unwrapped_values):
            return ElseIf(left, right)
        else:
            return Union(left, right)



from __future__ import annotations

import enum
import importlib
import inspect
import logging
import sys
from dataclasses import dataclass, Field
from datetime import datetime
from functools import cached_property, lru_cache
from types import NoneType
from collections.abc import Sequence

from typing_extensions import (
    get_type_hints,
    get_origin,
    get_args,
    ClassVar,
    List,
    Type,
    TYPE_CHECKING,
    Optional,
    Union,
)

from .utils import is_builtin_class
from ..ormatic.utils import module_and_class_name

if TYPE_CHECKING:
    from .class_diagram import WrappedClass


@dataclass
class TypeResolutionError(TypeError):
    """
    Error raised when a type cannot be resolved, even if searched for manually.
    """

    name: str

    def __post_init__(self):
        super().__init__(f"Could not resolve type for {self.name}")


@dataclass
class WrappedField:
    """
    A class that wraps a field of dataclass and provides some utility functions.
    """

    clazz: WrappedClass
    """
    The wrapped class that the field was created from.
    """

    field: Field
    """
    The dataclass field object that is wrapped.
    """

    container_types: ClassVar[List[Type]] = [list, set, tuple, type, Sequence]
    """
    A list of container types that are supported by the parser.
    """

    def __hash__(self):
        return hash((self.clazz.clazz, self.field))

    def __eq__(self, other):
        return (self.clazz.clazz, self.field) == (
            other.clazz.clazz,
            other.field,
        )

    def __repr__(self):
        return f"{module_and_class_name(self.clazz.clazz)}.{self.field.name}"

    @cached_property
    def resolved_type(self):
        try:
            result = get_type_hints(self.clazz.clazz)[self.field.name]
        except NameError as e:
            found_clazz = manually_search_for_class_name(e.name)
            module = importlib.import_module(found_clazz.__module__)
            locals()[e.name] = getattr(module, e.name)
            result = get_type_hints(self.clazz.clazz, localns=locals())[self.field.name]
        return result

    @cached_property
    def is_builtin_type(self) -> bool:

        return self.type_endpoint in [int, float, str, bool, datetime, NoneType]

    @cached_property
    def is_container(self) -> bool:
        return get_origin(self.resolved_type) in self.container_types

    @cached_property
    def container_type(self) -> Optional[Type]:
        if not self.is_container:
            return None
        return get_origin(self.resolved_type)

    @cached_property
    def is_collection_of_builtins(self):
        return self.is_container and all(
            is_builtin_class(field_type) for field_type in get_args(self.resolved_type)
        )

    @cached_property
    def is_optional(self):
        origin = get_origin(self.resolved_type)
        if origin not in [Union, Optional]:
            return False
        if origin == Union:
            args = get_args(self.resolved_type)
            return len(args) == 2 and NoneType in args
        return True

    @cached_property
    def contained_type(self):
        if not self.is_container and not self.is_optional:
            raise ValueError("Field is not a container")
        if self.is_optional:
            return get_args(self.resolved_type)[0]
        else:
            try:
                return get_args(self.resolved_type)[0]
            except IndexError:
                if self.resolved_type is Type:
                    return self.resolved_type
                else:
                    raise

    @cached_property
    def is_type_type(self) -> bool:
        return get_origin(self.resolved_type) is type

    @cached_property
    def is_enum(self) -> bool:
        if self.is_container:
            return False
        if self.is_optional:
            return issubclass(self.contained_type, enum.Enum)

        return issubclass(self.resolved_type, enum.Enum)

    @cached_property
    def is_one_to_one_relationship(self) -> bool:
        return not self.is_container and not self.is_builtin_type

    @cached_property
    def is_one_to_many_relationship(self) -> bool:
        return self.is_container and not self.is_builtin_type and not self.is_optional

    @cached_property
    def type_endpoint(self) -> Type:
        if self.is_container or self.is_optional:
            return self.contained_type
        else:
            return self.resolved_type


@lru_cache(maxsize=None)
def manually_search_for_class_name(target_class_name: str) -> Type:
    """
    Searches for a class with the specified name in the current module's `globals()` dictionary
    and all loaded modules present in `sys.modules`. This function attempts to find and resolve
    the first class that matches the given name. If multiple classes are found with the same
    name, a warning is logged, and the first one is returned. If no matching class is found,
    an exception is raised.

    :param target_class_name: Name of the class to search for.
    :return: The resolved class with the matching name.

    :raises ValueError: Raised when no class with the specified name can be found.
    """
    found_classes = search_class_in_globals(target_class_name)
    found_classes += search_class_in_sys_modules(target_class_name)

    if len(found_classes) == 0:
        raise TypeResolutionError(target_class_name)
    elif len(found_classes) == 1:
        resolved_class = found_classes[0]
    else:
        logging.warning(
            f"Found multiple classes with name {target_class_name}. Found classes: {found_classes} "
        )
        resolved_class = found_classes[0]

    return resolved_class


def search_class_in_globals(target_class_name: str) -> List[Type]:
    """
    Searches for a class with the given name in the current module's globals.

    :param target_class_name: The name of the class to search for.
    :return: The resolved classes with the matching name.
    """
    return [
        obj
        for obj in globals().items()
        if inspect.isclass(obj) and obj.__name__ == target_class_name
    ]


def search_class_in_sys_modules(target_class_name: str) -> List[Type]:
    """
    Searches for a class with the given name in all loaded modules (via sys.modules).
    """
    found_classes = []
    for module_name, module in sys.modules.items():
        if module is None or not hasattr(module, "__dict__"):
            continue  # Skip built-in modules or modules without a __dict__

        for name, obj in module.__dict__.items():
            if inspect.isclass(obj) and obj.__name__ == target_class_name:
                # Avoid duplicates if a class is imported into multiple namespaces
                if (obj, f"from module '{module_name}'") not in found_classes:
                    found_classes.append(obj)
    return found_classes



from __future__ import annotations

import typing
from abc import ABC
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Dict, Optional, Iterable

from .cache_data import SeenSet, is_caching_enabled
from .conclusion import Conclusion
from .hashed_data import HashedIterable, HashedValue
from .rxnode import ColorLegend
from .symbolic import LogicalOperator, SymbolicExpression, ElseIf, Union as EQLUnion, Literal


@dataclass(eq=False)
class ConclusionSelector(LogicalOperator, ABC):
    """
    Base class for logical operators that may carry and select conclusions.

    Tracks whether certain conclusion-combinations were already produced so
    they are not duplicated across truth branches.
    """
    concluded_before: Dict[bool, SeenSet] = field(default_factory=lambda: {True: SeenSet(), False: SeenSet()},
                                                  init=False)

    def update_conclusion(self, output: Dict[int, HashedValue], conclusions: typing.Set[Conclusion]) -> None:
        if not conclusions:
            return
        required_vars = HashedIterable()
        for conclusion in conclusions:
            vars_ = conclusion._unique_variables_.filter(lambda v: not isinstance(v.value, Literal))
            required_vars.update(vars_)
        required_output = {k: v for k, v in output.items() if k in required_vars}
        if not self.concluded_before[not self._is_false_].check(required_output):
            self._conclusion_.update(conclusions)
            self.concluded_before[not self._is_false_].add(required_output)

    def _copy_expression_(self, postfix: str) -> SymbolicExpression:
        cp = super()._copy_expression_(postfix)
        cp.concluded_before = {True: SeenSet(), False: SeenSet()}
        return cp

    @property
    def _plot_color_(self) -> ColorLegend:
        return ColorLegend("ConclusionSelector", '#eded18')


@dataclass(eq=False)
class ExceptIf(ConclusionSelector):
    """
    Conditional branch that yields left unless the right side produces values.

    This encodes an "except if" behavior: when the right condition matches,
    the left branch's conclusions/outputs are excluded; otherwise, left flows through.
    """

    @lru_cache(maxsize=None)
    def _required_variables_from_child_(self, child: Optional[SymbolicExpression] = None, when_true: bool = True):
        if not child:
            child = self.left
        required_vars = HashedIterable()
        when_false = not when_true
        if child is self.left:
            if when_true:
                required_vars.update(self.right._unique_variables_)
            for conc in self.left._conclusion_.union(self.right._conclusion_):
                required_vars.update(conc._unique_variables_)
        elif child is self.right:
            if when_true:
                for conc in self.right._conclusion_:
                    required_vars.update(conc._unique_variables_)
            if when_false and not self.left._is_false_:
                for conc in self.left._conclusion_:
                    required_vars.update(conc._unique_variables_)
        required_vars.update(self._parent_._required_variables_from_child_(self, when_true))
        for conc in self._conclusion_:
            required_vars.update(conc._unique_variables_)
        return required_vars

    def _evaluate__(self, sources: Optional[Dict[int, HashedValue]] = None, yield_when_false: bool = False) -> Iterable[Dict[int, HashedValue]]:
        """
        Evaluate the ExceptIf condition and yield the results.
        """

        # init an empty source if none is provided
        sources = sources or HashedIterable()
        self._yield_when_false_ = yield_when_false

        # constrain left values by available sources
        left_values = self.left._evaluate__(sources, yield_when_false=self._yield_when_false_)
        for left_value in left_values:

            left_value.update(sources)

            self._is_false_ = self.left._is_false_
            if self._is_false_:
                if self._yield_when_false_:
                    if not self._is_duplicate_output_(left_value):
                        yield left_value
                continue

            if is_caching_enabled() and self.right_cache.check(left_value):
                yield from self.yield_final_output_from_cache(left_value, self.right_cache)
                continue

            right_yielded = False
            for right_value in self.right._evaluate__(left_value, yield_when_false=False):
                right_yielded = True
                self._conclusion_.update(self.right._conclusion_)
                output = left_value.copy()
                output.update(right_value)
                yield output
                self._conclusion_.clear()
            if not right_yielded:
                self._conclusion_.update(self.left._conclusion_)
                yield left_value
                self._conclusion_.clear()


@dataclass(eq=False)
class Alternative(ElseIf, ConclusionSelector):
    """
    A conditional branch that behaves like an "else if" clause where the left branch
    is selected if it is true, otherwise the right branch is selected if it is true else
    none of the branches are selected.
    """

    def _is_duplicate_output_(self, output: Dict[int, HashedValue]) -> bool:
        # For alternatives, avoid suppressing outputs based solely on variable values,
        # as different branches may yield different conclusions for the same bindings.
        # Let ConclusionSelector.update_conclusion handle deduplication of conclusions.
        return False

    def _evaluate__(self, sources: Optional[Dict[int, HashedValue]] = None, yield_when_false: bool = False) -> Iterable[Dict[int, HashedValue]]:
        outputs = super()._evaluate__(sources, yield_when_false=yield_when_false)
        for output in outputs:
            left_is_true = not self.left._is_false_
            right_is_true = not self.right._is_false_
            if left_is_true:
                self.update_conclusion(output, self.left._conclusion_)
            elif right_is_true:
                self.update_conclusion(output, self.right._conclusion_)
            yield output
            self._conclusion_.clear()


@dataclass(eq=False)
class Next(EQLUnion, ConclusionSelector):
    """
    A Union conclusion selector that always evaluates the left and right branches and combines their results.
    """

    def _evaluate__(self, sources: Optional[Dict[int, HashedValue]] = None, yield_when_false: bool = False) -> Iterable[Dict[int, HashedValue]]:
        outputs = super()._evaluate__(sources, yield_when_false=yield_when_false)
        for output in outputs:
            if self.left_evaluated:
                self.update_conclusion(output, self.left._conclusion_)
            if self.right_evaluated:
                self.update_conclusion(output, self.right._conclusion_)
            yield output
            self._conclusion_.clear()


from typing import Type, Iterable, List, Set, Tuple, Union, Optional


def get_range_types(type_hint: Type) -> Iterable[Type]:
    origin = getattr(type_hint, '__origin__', None)
    args = getattr(type_hint, '__args__', [])
    if not origin:
        yield type_hint
    elif is_container_type(origin):
        range_type = args[0]
        if is_container_type(range_type):
            raise ValueError("Nested containers are not supported")
        else:
            yield range_type
    elif is_optional_type(origin, args):
        for arg in args:
            if arg is not type(None):
                yield arg
        yield type(None)


def is_container_type(type_origin: type) -> bool:
    return type_origin in [list, set, tuple, List, Set, Tuple]


def is_optional_type(type_origin: type, args: List[Type]) -> bool:
    return (type_origin is Optional) or (type_origin is Union and type(None) in args)



from __future__ import annotations

"""
Custom exception types used by entity_query_language.
"""
from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from .symbolic import SymbolicExpression


class MultipleSolutionFound(Exception):
    """
    Raised when a query unexpectedly yields more than one solution where a single
    result was expected.

    :param first_val: The first solution encountered.
    :param second_val: The second solution encountered.
    """
    def __init__(self, first_val, second_val):
        super(MultipleSolutionFound, self).__init__(
            f"Multiple solutions found, the first two are {first_val}\n{second_val}"
        )


class NoSolutionFound(Exception):
    """
    Raised when a query does not yield any solution.
    """
    def __init__(self, expression: SymbolicExpression):
        super(NoSolutionFound, self).__init__(f"No solution found for expression {expression}")



from __future__ import annotations

import datetime
import inspect
import sys
from contextlib import suppress
from enum import Enum


import sqlalchemy
from sqlalchemy import Engine, text, MetaData
from typing_extensions import TypeVar, _SpecialForm, Type, List, Iterable, Union


class classproperty:
    """
    A decorator that allows a class method to be accessed as a property.
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, instance, owner):
        return self.fget(owner)


def classes_of_module(module) -> List[Type]:
    """
    Get all classes of a given module.

    :param module: The module to inspect.
    :return: All classes of the given module.
    """

    result = []
    for name, obj in inspect.getmembers(sys.modules[module.__name__]):
        if inspect.isclass(obj) and obj.__module__ == module.__name__:
            result.append(obj)
    return result


T = TypeVar("T")


def recursive_subclasses(cls: Type[T]) -> List[Type[T]]:
    """
    :param cls: The class.
    :return: A list of the classes subclasses.
    """
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in recursive_subclasses(s)
    ]


leaf_types = (int, float, str, Enum, datetime.datetime, bool)


def _drop_fk_constraints(engine: Engine, tables: Iterable[str]) -> None:
    """
    Drops foreign key constraints for the specified tables in the given engine.

    This function removes all foreign key constraints for the specified list
    of tables using the provided database engine. It supports multiple
    SQL dialects, including MySQL, PostgreSQL, SQLite, and others.

    :param engine: The SQLAlchemy Engine instance used to interact with
        the database.
    :param tables: An iterable of table names whose foreign key constraints
        need to be dropped.
    """
    insp = sqlalchemy.inspect(engine)
    dialect = engine.dialect.name.lower()

    with engine.begin() as conn:
        for table in tables:
            for fk in insp.get_foreign_keys(table):
                name = fk.get("name")
                if not name:  # unnamed FKs (e.g. SQLite)
                    continue

                if dialect.startswith("mysql"):
                    stmt = text(f"ALTER TABLE `{table}` DROP FOREIGN KEY `{name}`")
                else:  # PostgreSQL, SQLite, MSSQL, …
                    stmt = text(f'ALTER TABLE "{table}" DROP CONSTRAINT "{name}"')

                with suppress(Exception):
                    conn.execute(stmt)


def drop_database(engine: Engine) -> None:
    """
     Drops all tables in the given database engine. This function removes foreign key
     constraints and tables in reverse dependency order to ensure that proper
     dropping of objects occurs without conflict. For MySQL/MariaDB, foreign key
    checks are disabled temporarily during the process.

     This method differs from sqlalchemy `MetaData.drop_all <https://docs.sqlalchemy.org/en/20/core/metadata.html#sqlalchemy.schema.MetaData.drop_all>`_\ such that databases containing cyclic
     backreferences are also droppable.

     :param engine: The SQLAlchemy Engine instance connected to the target database
         where tables will be dropped.
     :type engine: Engine
     :return: None
    """
    metadata = MetaData()
    metadata.reflect(bind=engine)

    if not metadata.tables:
        return

    # 1. Drop FK constraints that would otherwise block table deletion.
    _drop_fk_constraints(engine, metadata.tables.keys())

    # 2. On MySQL / MariaDB it is still safest to disable FK checks entirely
    #    while the DROP TABLE statements run; other back-ends don’t need this.
    disable_fk_checks = engine.dialect.name.lower().startswith("mysql")

    with engine.begin() as conn:
        if disable_fk_checks:
            conn.execute(text("SET FOREIGN_KEY_CHECKS = 0"))

        # Drop in reverse dependency order (children first → parents last).
        for table in reversed(metadata.sorted_tables):
            table.drop(bind=conn, checkfirst=True)

        if disable_fk_checks:
            conn.execute(text("SET FOREIGN_KEY_CHECKS = 1"))


class InheritanceStrategy(Enum):
    JOINED = "joined"
    SINGLE = "single"


def module_and_class_name(t: Union[Type, _SpecialForm]) -> str:
    return f"{t.__module__}.{t.__name__}"



from __future__ import annotations

import logging
import os
import subprocess
import sys
from dataclasses import dataclass, field
from typing import TextIO, TYPE_CHECKING

import jinja2

if TYPE_CHECKING:
    from .ormatic import ORMatic

logger = logging.getLogger(__name__)


@dataclass
class SQLAlchemyGenerator:
    """
    A class for generating SQLAlchemy declarative mappings from ORMatic models.
    This class writes the tables generated by ORMatic to a file that serves as an SQL database interface.
    Uses Jinja2 templates for code generation.
    """

    ormatic: ORMatic
    """
    The ORMatic instance that created this SQLAlchemyGenerator.
    """

    env: jinja2.Environment = field(init=False, default=None)
    """
    The environment to use with jinja2.
    """

    def __post_init__(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates")
        self.env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def to_sqlalchemy_file(self, file: TextIO):
        """
        Generate a Python file with SQLAlchemy declarative mappings from the ORMatic models.

        :param file: The file to write to
        """
        # Load the template
        template = self.env.get_template("sqlalchemy_model.py.jinja")

        # Render the template
        output = template.render(
            ormatic=self.ormatic,
        )

        # Write the output to the file
        file.write(output)

        # format the output with black
        command = [sys.executable, "-m", "black", str(file.name)]
        subprocess.run(command, capture_output=True, text=True, check=True)



from __future__ import annotations

import os
from dataclasses import dataclass, field, fields
from functools import cached_property
from typing_extensions import TYPE_CHECKING, Any, Iterable, Optional, List, Type


from rustworkx import PyDiGraph

from .. import logger
from ..class_diagrams import ClassDiagram, Relation
from ..class_diagrams.wrapped_field import WrappedField

if TYPE_CHECKING:
    from .predicate import Predicate, Symbol


@dataclass
class PredicateRelation(Relation):
    source: WrappedInstance
    target: WrappedInstance
    predicate: Predicate
    inferred: bool = False

    def __str__(self):
        return self.predicate.__class__.__name__

    @property
    def color(self) -> str:
        return "red" if self.inferred else "black"


@dataclass
class WrappedInstance:
    instance: Symbol
    index: Optional[int] = field(init=False, default=None)
    _symbol_graph_: Optional[SymbolGraph] = field(
        init=False, hash=False, default=None, repr=False
    )
    inferred: bool = False

    @cached_property
    def fields(self) -> List[WrappedField]:
        return [WrappedField(self.instance, f) for f in fields(self.instance)]

    @property
    def name(self):
        return self.instance.__class__.__name__ + str(self.index)

    @property
    def color(self) -> str:
        return "red" if self.inferred else "black"

    def __eq__(self, other):
        return self.instance == other.instance

    def __hash__(self):
        try:
            return hash(self.instance)
        except TypeError:
            return id(self.instance)


@dataclass
class SymbolGraph:
    """
    A more encompassing class diagram that includes relations between classes other than inheritance and associations.
    Relations are represented as edges where each edge has a relation object attached to it. The relation object
    contains also the Predicate object that represents the relation.
    """

    _type_graph: ClassDiagram
    _instance_graph: PyDiGraph[WrappedInstance, PredicateRelation] = field(
        default_factory=PyDiGraph
    )

    def add_node(self, wrapped_instance: WrappedInstance) -> None:
        if not isinstance(wrapped_instance, WrappedInstance):
            wrapped_instance = WrappedInstance(wrapped_instance)
        wrapped_instance.index = self._instance_graph.add_node(wrapped_instance)
        wrapped_instance._symbol_graph_ = self

    # Adapters to align with ORM alternative mapping expectations
    def add_instance(self, wrapped_instance: WrappedInstance) -> None:
        """Add a wrapped instance to the graph.

        This is an adapter that delegates to add_node to keep API compatibility with
        SymbolGraphMapping.create_from_dao.
        """
        self.add_node(wrapped_instance)

    def add_relation(self, relation: PredicateRelation) -> None:
        """Add a relation edge to the instance graph.

        This is an adapter that delegates to add_edge to keep API compatibility with
        SymbolGraphMapping.create_from_dao.
        """
        self.add_edge(relation)

    def add_edge(self, relation: PredicateRelation) -> None:
        source_out_edges = self._instance_graph.out_edges(relation.source.index)
        for _, child_idx, e in source_out_edges:
            if (
                type(e.predicate) == type(relation.predicate)
                and child_idx == relation.target.index
            ):
                return
        self._instance_graph.add_edge(
            relation.source.index, relation.target.index, relation
        )

    def relations(self) -> Iterable[PredicateRelation]:
        yield from self._instance_graph.edges()

    @property
    def wrapped_instances(self) -> List[WrappedInstance]:
        return self._instance_graph.nodes()

    def get_wrapped_instance(self, instance: Any) -> Optional[WrappedInstance]:
        for wrapped_instance in self.wrapped_instances:
            if wrapped_instance.instance is instance:
                return wrapped_instance
        return None

    def get_outgoing_neighbors_with_edge_type(
        self,
        wrapped_instance: WrappedInstance,
        predicate_type: Type[Predicate],
    ) -> Iterable[WrappedInstance]:
        for _, child_idx, edge in self._instance_graph.out_edges(
            wrapped_instance.index
        ):
            if isinstance(edge.predicate, predicate_type):
                yield self._instance_graph.get_node_data(child_idx)

    def get_outgoing_neighbors(
        self, wrapped_instance: WrappedInstance
    ) -> Iterable[WrappedInstance]:
        yield from (
            self._instance_graph.get_node_data(child_idx)
            for _, child_idx, _ in self._instance_graph.out_edges(
                wrapped_instance.index
            )
        )

    def get_neighbors(
        self, wrapped_instance: WrappedInstance
    ) -> Iterable[WrappedInstance]:
        yield from (
            self._instance_graph.get_node_data(idx)
            for idx in self._instance_graph.neighbors(wrapped_instance.index)
        )

    def clear(self):
        self._type_graph.clear()
        self._instance_graph.clear()

    def to_dot(
        self,
        filepath: str,
        format="svg",
        graph_type="instance",
        without_inherited_associations: bool = True,
    ) -> None:
        import pydot

        if graph_type == "type":
            if without_inherited_associations:
                graph = (
                    self._type_graph.to_subdiagram_without_inherited_associations()._dependency_graph
                )
            else:
                graph = self._type_graph._dependency_graph
        else:
            graph = self._instance_graph
        if not filepath.endswith(f".{format}"):
            filepath += f".{format}"
        dot_str = graph.to_dot(
            lambda node: dict(
                color="black",
                fillcolor="lightblue",
                style="filled",
                label=node.name,
            ),
            lambda edge: dict(color=edge.color, style="solid", label=str(edge)),
            dict(rankdir="LR"),
        )
        dot = pydot.graph_from_dot_data(dot_str)[0]
        try:
            dot.write(filepath, format=format)
        except FileNotFoundError:
            tmp_filepath = filepath.replace(f".{format}", ".dot")
            dot.write(tmp_filepath, format="raw")
            try:
                os.system(f"/usr/bin/dot -T{format} {tmp_filepath} -o {filepath}")
                os.remove(tmp_filepath)
            except Exception as e:
                logger.error(e)



__version__ = "1.1.16"

import logging
import sys

# Configure default logging for all loggers in the package
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(
    logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
)

# Get the package logger
logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.INFO)

# Configure all module loggers
for module_name in ["dao", "ormatic", "sqlalchemy_generator"]:
    module_logger = logging.getLogger(f"{__name__}.{module_name}")
    module_logger.addHandler(handler)
    module_logger.setLevel(logging.INFO)



import inspect
import sys
from typing import List, Type


def classes_of_module(module) -> List[Type]:
    """
    Get all classes of a given module.

    :param module: The module to inspect.
    :return: All classes of the given module.
    """

    result = []
    for name, obj in inspect.getmembers(sys.modules[module.__name__]):
        if inspect.isclass(obj) and obj.__module__ == module.__name__:
            result.append(obj)
    return result


def is_builtin_class(clazz: Type) -> bool:
    return clazz.__module__ == "builtins"



from __future__ import annotations

from dataclasses import dataclass
from typing import Type, Iterable, Tuple, Union

"""
Utilities for hashing, rendering, and general helpers used by the
symbolic query engine.
"""
import codecs
import itertools
import os
import re
from subprocess import check_call
from tempfile import NamedTemporaryFile

try:
    import six
except ImportError:
    six = None

try:
    from graphviz import Source
except ImportError:
    Source = None

from typing_extensions import Callable, Set, Any, Optional, List


class IDGenerator:
    """
    A class that generates incrementing, unique IDs and caches them for every object this is called on.
    """

    _counter = 0
    """
    The counter of the unique IDs.
    """

    # @lru_cache(maxsize=None)
    def __call__(self, obj: Any) -> int:
        """
        Creates a unique ID and caches it for every object this is called on.

        :param obj: The object to generate a unique ID for, must be hashable.
        :return: The unique ID.
        """
        self._counter += 1
        return self._counter


def lazy_iterate_dicts(dict_of_iterables):
    """Generator that yields dicts with one value from each iterable"""
    for values in zip(*dict_of_iterables.values()):
        yield dict(zip(dict_of_iterables.keys(), values))


def generate_combinations(generators_dict):
    """Yield all combinations of generator values as keyword arguments"""
    for combination in itertools.product(*generators_dict.values()):
        yield dict(zip(generators_dict.keys(), combination))


def filter_data(data, selected_indices):
    data = iter(data)
    prev = -1
    encountered_indices = set()
    for idx in selected_indices:
        if idx in encountered_indices:
            continue
        encountered_indices.add(idx)
        skip = idx - prev - 1
        data = itertools.islice(data, skip, None)
        try:
            yield next(data)
        except StopIteration:
            break
        prev = idx


def make_list(value: Any) -> List:
    """
    Make a list from a value.

    :param value: The value to make a list from.
    """
    return list(value) if is_iterable(value) else [value]


def is_iterable(obj: Any) -> bool:
    """
    Check if an object is iterable.

    :param obj: The object to check.
    """
    return hasattr(obj, "__iter__") and not isinstance(
        obj, (str, type, bytes, bytearray)
    )


def make_tuple(value: Any) -> Any:
    """
    Make a tuple from a value.
    """
    return tuple(value) if is_iterable(value) else (value,)


def make_set(value: Any) -> Set:
    """
    Make a set from a value.

    :param value: The value to make a set from.
    """
    return set(value) if is_iterable(value) else {value}


@dataclass(eq=False)
class ALL:
    """
    Sentinel that compares equal to any other value.

    This is used to signal wildcard matches in hashing/containment logic.
    """

    def __eq__(self, other):
        """Always return True."""
        return True

    def __hash__(self):
        """Hash based on object identity to remain unique as a sentinel."""
        return hash(id(self))


All = ALL()



from __future__ import annotations

import logging
from abc import ABC
from dataclasses import dataclass
from dataclasses import field, InitVar, fields
from functools import cached_property
from typing import List, Optional

import rustworkx as rx
from rustworkx_utils import RWXNode
from typing_extensions import Type

from krrood.class_diagrams.wrapped_field import WrappedField


@dataclass
class Relation(ABC):
    """
    Abstract base class representing a relationship between two classes in a UML class diagram.
    """

    source: WrappedClass
    """The source class in the relation."""

    target: WrappedClass
    """The target class in the relation."""

    def __str__(self):
        return f"{self.__class__.__name__}"

    @property
    def color(self) -> str:
        return "black"


@dataclass
class Inheritance(Relation):
    """
    Represents an inheritance (generalization) relationship in UML.

    This is an "is-a" relationship where the source class inherits from the target class.
    In UML notation, this is represented by a solid line with a hollow triangle pointing to the parent class.
    """

    def __str__(self):
        return f"isSuperClassOf"


@dataclass
class Association(Relation):
    """
    Represents a general association relationship between two classes.

    This is the most general form of relationship, indicating that instances of one class
    are connected to instances of another class. In UML notation, this is shown as a solid line.
    """

    field: WrappedField
    """The field in the source class that creates this association with the target class."""

    def __str__(self):
        return f"has-{self.field.field.name}"


class ParseError(TypeError):
    """
    Error that will be raised when the parser encounters something that can/should not be parsed.

    For instance, Union types
    """

    pass


@dataclass
class WrappedClass:
    index: Optional[int] = field(init=False, default=None)
    clazz: Type
    _class_diagram: Optional[ClassDiagram] = field(
        init=False, hash=False, default=None, repr=False
    )

    @cached_property
    def fields(self) -> List[WrappedField]:
        try:
            return [
                WrappedField(self, f)
                for f in fields(self.clazz)
                if not f.name.startswith("_")
            ]
        except TypeError as e:
            logging.error(f"Error parsing class {self.clazz}: {e}")
            raise ParseError(e) from e

    @property
    def name(self):
        return self.clazz.__name__ + str(self.index)

    def __hash__(self):
        return hash((self.index, self.clazz))


@dataclass
class ClassDiagram:

    classes: InitVar[List[Type]]

    _dependency_graph: rx.PyDiGraph[WrappedClass, Relation] = field(
        default_factory=rx.PyDiGraph, init=False
    )

    def __post_init__(self, classes: List[Type]):
        self._dependency_graph = rx.PyDiGraph()
        for clazz in classes:
            self.add_node(WrappedClass(clazz=clazz))
        self._create_all_relations()

    def to_subdiagram_without_inherited_associations(
        self,
        include_field_name: bool = False,
    ) -> ClassDiagram:
        """
        Return a new class diagram where association edges that are present on any
        ancestor of the source class are removed from descendants.

        Inheritance edges are preserved.
        """
        # Rebuild a fresh diagram from the same classes to avoid mutating this instance
        result = ClassDiagram(classes=[wc.clazz for wc in self.wrapped_classes])

        # Convenience locals
        g = result._dependency_graph

        # 1) Build parent map from inheritance edges: child_idx -> set(parent_idx)
        parent_map: dict[int, set[int]] = {}
        for u, v in g.edge_list():
            rel = g.get_edge_data(u, v)
            if isinstance(rel, Inheritance):
                parent_map.setdefault(v, set()).add(u)

        # 2) DFS to compute all ancestors for each node index
        def all_ancestors(node_idx: int) -> set[int]:
            parents = parent_map.get(node_idx, set())
            if not parents:
                return set()
            stack = list(parents)
            seen: set[int] = set(parents)
            while stack:
                cur = stack.pop()
                for p in parent_map.get(cur, set()):
                    if p not in seen:
                        seen.add(p)
                        stack.append(p)
            return seen

        # 3) Precompute association keys per source node
        #    Key = (relation class, target class[, field name])
        def assoc_key(rel: Association) -> tuple:
            if include_field_name:
                return (rel.__class__, rel.target.clazz, rel.field.field.name)
            return (rel.__class__, rel.target.clazz)

        assoc_keys_by_source: dict[int, set[tuple]] = {}
        for u, v in g.edge_list():
            rel = g.get_edge_data(u, v)
            if isinstance(rel, Association):
                assoc_keys_by_source.setdefault(u, set()).add(assoc_key(rel))

        # 4) Mark redundant descendant association edges for removal
        edges_to_remove: list[tuple[int, int]] = []
        for u, v in g.edge_list():
            rel = g.get_edge_data(u, v)
            if not isinstance(rel, Association):
                continue

            key = assoc_key(rel)
            # Collect all keys defined by any ancestor of u
            inherited_keys: set[tuple] = set()
            for anc in all_ancestors(u):
                inherited_keys |= assoc_keys_by_source.get(anc, set())

            if key in inherited_keys:
                edges_to_remove.append((u, v))

        # 5) Remove redundant edges
        for u, v in edges_to_remove:
            # Safe even if duplicates appear in list; graph ignores missing
            try:
                g.remove_edge(u, v)
            except Exception:
                # Be conservative: if already removed due to earlier operation, skip
                pass

        return result

    @property
    def wrapped_classes(self):
        return self._dependency_graph.nodes()

    @property
    def associations(self) -> List[Association]:
        return [
            edge
            for edge in self._dependency_graph.edges()
            if isinstance(edge, Association)
        ]

    @property
    def inheritance_relations(self) -> List[Inheritance]:
        return [
            edge
            for edge in self._dependency_graph.edges()
            if isinstance(edge, Inheritance)
        ]

    def get_wrapped_class(self, clazz: Type) -> Optional[WrappedClass]:
        base = [cls for cls in self.wrapped_classes if cls.clazz == clazz]
        return base[0] if base else None

    def add_node(self, clazz: WrappedClass):
        clazz.index = self._dependency_graph.add_node(clazz)
        clazz._class_diagram = self

    def add_relation(self, relation: Relation):
        self._dependency_graph.add_edge(
            relation.source.index, relation.target.index, relation
        )

    def _create_inheritance_relations(self):
        for clazz in self.wrapped_classes:
            for superclass in clazz.clazz.__bases__:
                source = self.get_wrapped_class(superclass)
                if source:
                    relation = Inheritance(
                        source=source,
                        target=clazz,
                    )
                    self.add_relation(relation)

    def _create_all_relations(self):
        self._create_inheritance_relations()
        self._create_association_relations()

    def _create_association_relations(self):
        for clazz in self.wrapped_classes:
            for wrapped_field in clazz.fields:
                target_type = (
                    wrapped_field.contained_type
                    if wrapped_field.is_container or wrapped_field.is_optional
                    else wrapped_field.resolved_type
                )

                wrapped_target_class = self.get_wrapped_class(target_type)

                if not wrapped_target_class:
                    continue

                relation = Association(
                    field=wrapped_field,
                    source=clazz,
                    target=wrapped_target_class,
                )
                self.add_relation(relation)

    def _build_rxnode_tree(self) -> RWXNode:
        """
        Convert the class diagram graph to RWXNode tree structure for visualization.

        Creates a tree where inheritance relationships are represented as parent-child connections.
        If there are multiple root classes, they are grouped under a virtual root node.

        :return: Root RWXNode representing the class diagram
        """
        # Create RWXNode for each class
        node_map = {}
        for wrapped_class in self.wrapped_classes:
            class_name = wrapped_class.clazz.__name__
            node = RWXNode(name=class_name, data=wrapped_class)
            node_map[wrapped_class.index] = node

        # Build parent-child relationships from edges
        for edge in self._dependency_graph.edge_list():
            source_idx, target_idx = edge
            relation = self._dependency_graph.get_edge_data(source_idx, target_idx)

            # For inheritance: source is parent class, target is child class
            # In RWXNode: parent class should have child class as its child
            if isinstance(relation, Inheritance):
                parent_node = node_map[source_idx]
                child_node = node_map[target_idx]
                child_node.add_parent(parent_node)
            # elif isinstance(relation, Association):
            #     # For associations, add as parent relationship with label
            #     source_node = node_map[source_idx]
            #     target_node = node_map[target_idx]
            #     # Association goes from source to target
            #     target_node.add_parent(source_node)

        # Find root nodes (nodes without parents)
        root_nodes = [node for node in node_map.values() if not node.parents]

        # If there's only one root, return it
        if len(root_nodes) == 1:
            return root_nodes[0]

        # If there are multiple roots, create a virtual root
        virtual_root = RWXNode(name="Class Diagram")
        for root_node in root_nodes:
            root_node.add_parent(virtual_root)

        return virtual_root

    def visualize(
        self,
        filename: str = "class_diagram.pdf",
        title: str = "Class Diagram",
        figsize: tuple = (35, 30),
        node_size: int = 7000,
        font_size: int = 25,
        layout: str = "layered",
        edge_style: str = "straight",
        **kwargs,
    ):
        """
        Visualize the class diagram using rustworkx_utils.

        Creates a visual representation of the class diagram showing classes and their relationships.
        The diagram is saved as a PDF file.

        :param filename: Output filename for the visualization
        :param title: Title for the diagram
        :param figsize: Figure size as (width, height) tuple
        :param node_size: Size of the nodes in the visualization
        :param font_size: Font size for labels
        :param kwargs: Additional keyword arguments passed to RWXNode.visualize()
        """
        root_node = self._build_rxnode_tree()
        root_node.visualize(
            filename=filename,
            title=title,
            figsize=figsize,
            node_size=node_size,
            font_size=font_size,
            layout=layout,
            edge_style=edge_style,
            **kwargs,
        )

    def clear(self):
        self._dependency_graph.clear()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Generic, Optional, Iterable, Dict, Any, Callable, List, Union

from typing_extensions import TypeVar

from .utils import make_list, ALL

T = TypeVar("T")


@dataclass
class HashedValue(Generic[T]):
    """
    Value wrapper carrying a stable hash identifier.

    :param value: The wrapped value.
    :param id_: Optional explicit identifier; if omitted, derived from value.
    :ivar value: The wrapped value.
    :ivar id_: The stable identifier used for hashing and equality.
    """

    value: T
    id_: Optional[int] = field(default=None)

    def __post_init__(self) -> None:
        """
        Initialize the identifier from the wrapped value when not provided.
        """
        if self.id_ is None:
            if isinstance(self.value, HashedValue):
                self.id_ = self.value.id_
                self.value = self.value.value
            elif hasattr(self.value, "_id_"):
                self.id_ = self.value._id_
            else:
                self.id_ = id(self.value)

    def __hash__(self) -> int:
        """Hash of the identifier."""
        return hash(self.id_)

    def __eq__(self, other: object) -> bool:
        """
        Equality based on identifier, with ALL sentinel matching any value.
        """
        if isinstance(other, ALL):
            return True
        if not isinstance(other, HashedValue):
            return False
        return self.id_ == other.id_


@dataclass
class HashedIterable(Generic[T]):
    """
    A wrapper for an iterable that hashes its items.
    This is useful for ensuring that the items in the iterable are unique and can be used as keys in a dictionary.
    """

    iterable: Iterable[HashedValue[T]] = field(default_factory=list)
    values: Dict[int, HashedValue[T]] = field(default_factory=dict)

    def __post_init__(self):
        if self.iterable and not isinstance(self.iterable, HashedIterable):
            self.iterable = (
                HashedValue(v) if not isinstance(v, HashedValue) else v
                for v in self.iterable
            )

    def set_iterable(self, iterable):
        if iterable and not isinstance(iterable, HashedIterable):
            self.iterable = (
                HashedValue(v) if not isinstance(v, HashedValue) else v
                for v in iterable
            )

    def get(self, key: int, default: Any) -> HashedValue[T]:
        return self.values.get(key, default)

    def add(self, value: Any):
        if not isinstance(value, HashedValue):
            value = HashedValue(value)
        if value.id_ not in self.values:
            self.values[value.id_] = value
        return self

    def update(self, iterable: Iterable[Any]):
        for v in iterable:
            self.add(v)

    def map(
        self,
        func: Callable[[HashedValue], HashedValue],
        ids: Optional[List[int]] = None,
    ) -> HashedIterable[T]:
        if ids:
            func = lambda v: func(v) if v.id_ in ids else v
        return HashedIterable(map(func, self))

    def filter(self, func: Callable[[HashedValue], bool]) -> HashedIterable[T]:
        return HashedIterable(filter(func, self))

    @property
    def unwrapped_values(self) -> List[T]:
        return [v.value for v in self]

    @property
    def first_value(self) -> HashedValue:
        """
        Return the first value in the iterable.
        """
        for v in self:
            return v
        raise ValueError("Tried to get a value from empty iterable")

    def clear(self):
        self.values.clear()

    def __iter__(self):
        """
        Iterate over the hashed values.

        :return: An iterator over the hashed values.
        """
        yield from self.values.values()
        for v in self.iterable:
            self.values[v.id_] = v
            yield v

    def __or__(self, other) -> HashedIterable[T]:
        return self.union(other)

    def __and__(self, other) -> HashedIterable[T]:
        return self.intersection(other)

    def intersection(self, other):
        common_keys = self.values.keys() & other.values.keys()
        common_values = {k: self.values[k] for k in common_keys}
        return HashedIterable(values=common_values)

    def difference(self, other):
        left_keys = self.values.keys() - other.values.keys()
        values = {k: self.values[k] for k in left_keys}
        return HashedIterable(values=values)

    def union(self, other):
        if not isinstance(other, HashedIterable):
            other = HashedIterable(
                values={HashedValue(v).id_: HashedValue(v) for v in make_list(other)}
            )
        all_keys = self.values.keys() | other.values.keys()
        all_values = {k: self.values.get(k, other.values.get(k)) for k in all_keys}
        return HashedIterable(values=all_values)

    def __len__(self) -> int:
        return len(self.values)

    def __getitem__(self, id_: Any) -> HashedValue:
        """
        Get the HashedValue by its id.

        :param id_: The id of the HashedValue to get.
        :return: The HashedValue with the given id.
        :raises KeyError: If the given id is unknown.
        """
        if isinstance(id_, HashedValue):
            id_ = id_.id_
        elif not isinstance(id_, int):
            id_ = HashedValue(id_).id_
        try:
            return self.values[id_]
        except KeyError:
            for v in self.iterable:
                self.values[v.id_] = v
                if v.id_ == id_:
                    return v
            raise KeyError(id_)

    def __setitem__(self, id_: int, value: HashedValue[T]):
        """
        Set the HashedValue by its id.

        :param id_: The id of the HashedValue to set.
        :param value: The HashedValue to set.
        """
        self.values[id_] = value

    def __copy__(self):
        """
        Create a shallow copy of the HashedIterable.

        :return: A new HashedIterable instance with the same values.
        """
        return HashedIterable(values=self.values.copy())

    def __contains__(self, item):
        return item in self.values

    def __hash__(self):
        return hash(tuple(sorted(self.values.keys())))

    def __eq__(self, other):
        keys_are_equal = self.values.keys() == other.values.keys()
        if not keys_are_equal:
            return False
        values_are_equal = all(
            my_v == other_v
            for my_v, other_v in zip(self.values.values(), other.values.values())
        )
        return values_are_equal

    def __bool__(self):
        return bool(self.values) or bool(self.iterable)
